#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Communication protocol for VISA. Makes use of the pyvisa library.
The backend can be NI-Visa or pyvisa-py.

Information on how to install a VISA backend can be found here:
https://pyvisa.readthedocs.io/en/master/getting_nivisa.html

So far only TCPIP SOCKET and TCPIP INSTR interfaces are supported.
"""

import logging
from collections.abc import Callable
from ipaddress import IPv4Address, IPv6Address
from time import sleep
from typing import TYPE_CHECKING, cast

import pyvisa as visa
from pyvisa_py.protocols.rpc import (  # type: ignore[import-untyped,import-not-found]
    RPCError,
)

from hvl_ccb.comm import CommunicationError
from hvl_ccb.comm.base import SyncCommunicationProtocol, SyncCommunicationProtocolConfig
from hvl_ccb.configuration import configdataclass
from hvl_ccb.utils.enum import AutoNumberNameEnum
from hvl_ccb.utils.typing import Number
from hvl_ccb.utils.validation import validate_and_resolve_host, validate_tcp_port

if TYPE_CHECKING:
    from pyvisa.resources import MessageBasedResource


logger = logging.getLogger(__name__)


class VisaCommunicationError(IOError, CommunicationError):
    """
    Base class for VisaCommunication errors.
    """


@configdataclass
class VisaCommunicationConfig(SyncCommunicationProtocolConfig):
    """
    `VisaCommunication` configuration dataclass.
    """

    class InterfaceType(AutoNumberNameEnum, init="value"):  # type: ignore[call-arg]
        """
        Supported VISA Interface types.
        """

        #: VISA-RAW protocol
        TCPIP_SOCKET = ((),)

        #: VXI-11 protocol
        TCPIP_INSTR = ((),)

        def address(
            self, host: str, port: int | None = None, board: int | None = None
        ) -> str:
            """
            Address string specific to the VISA interface type.

            :param host: host IP address
            :param port: optional TCP port
            :param board: optional board number
            :return: address string
            """

            if self.name == VisaCommunicationConfig.InterfaceType.TCPIP_SOCKET:
                return f"TCPIP{board}::{host}::{port}::SOCKET"
            if self.name == VisaCommunicationConfig.InterfaceType.TCPIP_INSTR:
                return f"TCPIP::{host}::INSTR"
            msg = f"{self.name} is unknown InterfaceType."
            raise VisaCommunicationError(msg)

    # IP address of the VISA device.
    host: str | IPv4Address | IPv6Address = ""

    #: Interface type of the VISA connection, being one of :class:`InterfaceType`.
    interface_type: str | InterfaceType = ""

    #: Board number is typically 0 and comes from old bus systems.
    board: int = 0

    #: TCP port, standard is 5025.
    port: int = 5025

    #: Timeout for commands in milli seconds.
    timeout: int = 5000

    #: Chunk size is the allocated memory for read operations. The standard is 20kB,
    #: and is increased per default here to 200kB. It is specified in bytes.
    chunk_size: int = 204800

    #: Timeout for opening the connection, in milli seconds.
    open_timeout: int = 1000

    #: The terminator will be used to set `write_termination` and `read_termination`
    terminator: bytes = b"\n"

    visa_backend: str = ""
    """
    Specifies the path to the library to be used with PyVISA as a backend. Defaults
    to None, which is NI-VISA (if installed), or pyvisa-py (if NI-VISA is not found).
    To force the use of pyvisa-py, specify '@py' here.
    """

    def clean_values(self) -> None:
        # in principle, host is allowed to be IP or FQDN. However, we only allow IP:
        self.force_value("host", validate_and_resolve_host(self.host, logger))  # type: ignore[attr-defined]
        validate_tcp_port(self.port, logger)

        if not isinstance(self.interface_type, self.InterfaceType):
            self.force_value("interface_type", self.InterfaceType(self.interface_type))  # type: ignore[attr-defined]

        if self.board < 0:
            msg = "Board number has to be >= 0."
            raise ValueError(msg)

        if self.timeout < 0:
            msg = "Timeout has to be >= 0."
            raise ValueError(msg)

        if self.open_timeout < 0:
            msg = "Open Timeout has to be >= 0."
            raise ValueError(msg)

        allowed_terminators = (b"\n", b"\r", b"\r\n")
        if self.terminator not in allowed_terminators:
            msg = "Terminator has to be \\n, \\r or \\r\\n as `byte`."
            raise ValueError(msg)

    @property
    def address(self) -> str:
        """
        Address string depending on the VISA protocol's configuration.

        :return: address string corresponding to current configuration
        """

        return self.interface_type.address(  # type: ignore[union-attr]
            cast("str", self.host),
            port=self.port,
            board=self.board,
        )


def has_instrument(func: Callable) -> Callable:
    """
    Decorator to check if `self._instrument is not None`.
    """

    def decorator(*args, **kwargs):
        self = args[0]
        if self._instrument is None:
            msg = (
                f'Could not execute "{func.__name__}" of VISA connection, it was not'
                " started."
            )
            logger.error(msg)
            raise VisaCommunicationError(msg)
        return func(*args, **kwargs)

    return decorator


class VisaCommunication(SyncCommunicationProtocol):
    """
    Implements the Communication Protocol for VISA / SCPI.
    """

    #: The maximum of commands that can be sent in one round is 5 according to the
    #: VISA standard.
    MULTI_COMMANDS_MAX = 5

    #: The character to separate two commands is ; according to the VISA standard.
    MULTI_COMMANDS_SEPARATOR = ";"

    #: Small pause in seconds to wait after write operations, allowing devices to
    #: really do what we tell them before continuing with further tasks.
    WAIT_AFTER_WRITE = 0.08  # seconds to wait after a write is sent

    def __init__(self, configuration) -> None:
        """
        Constructor for VisaCommunication.
        """

        super().__init__(configuration)

        # create a new resource manager
        if self.config.visa_backend == "":
            self._resource_manager = visa.ResourceManager()
        else:
            self._resource_manager = visa.ResourceManager(self.config.visa_backend)

        self._instrument: MessageBasedResource | None = None

    @staticmethod
    def config_cls() -> type[VisaCommunicationConfig]:
        return VisaCommunicationConfig

    def open(self) -> None:
        """
        Open the VISA connection and create the resource.
        """

        logger.info("Open the VISA connection.")

        with self.access_lock:
            try:
                self._instrument = cast(
                    "MessageBasedResource",
                    self._resource_manager.open_resource(
                        self.config.address,
                        open_timeout=self.config.open_timeout,
                    ),
                )
                self._instrument.chunk_size = self.config.chunk_size
                self._instrument.timeout = self.config.timeout
                self._instrument.write_termination = self.config.terminator.decode()
                self._instrument.read_termination = self.config.terminator.decode()

            except visa.VisaIOError as e:
                logger.exception("Error of VISA Communication", exc_info=e)
                if e.error_code != 0:
                    raise VisaCommunicationError from e

            except (
                RPCError,
                ConnectionRefusedError,
                BrokenPipeError,
            ) as e:
                # if pyvisa-py is used as backend, this RPCError can come. As it is
                # difficult to import (hyphen in package name), we "convert" it here to
                # a VisaCommunicationError. Apparently on the Linux runners,
                # a ConnectionRefusedError is raised on fail, rather than an RPCError.
                # On macOS the BrokenPipeError error is raised (from
                # pyvisa-py/protocols/rpc.py:320), with puzzling log message from
                # visa.py: "187 WARNING  Could not close VISA connection, was not
                # started."
                logger.exception("Error of VISA Communication", exc_info=e)
                raise VisaCommunicationError from e

            if self._instrument is not None:
                try:
                    # enable keep-alive of the connection. Seems not to work always, but
                    # using the status poller a keepalive of the connection is also
                    # satisfied. Unsupported on RTO 1022 devices.
                    value = visa.constants.VI_TRUE
                    if isinstance(
                        self._instrument.get_visa_attribute(
                            visa.constants.ResourceAttribute.tcpip_keepalive
                        ),
                        bool,
                    ):
                        value = True
                    self._instrument.set_visa_attribute(
                        visa.constants.ResourceAttribute.tcpip_keepalive, value
                    )

                except visa.VisaIOError as e:
                    if e.abbreviation == "VI_ERROR_NSUP_ATTR":
                        logger.warning(
                            "Error of VISA Communication: 'VI_ERROR_NSUP_ATTR'",
                            exc_info=e,
                        )
                    else:
                        logger.exception("Error of VISA Communication", exc_info=e)
                        if e.error_code != 0:
                            raise VisaCommunicationError from e

    def close(self) -> None:
        """
        Close the VISA connection and invalidates the handle.
        """

        if self._instrument is None:
            logger.warning("Could not close VISA connection, was not started.")
            return

        try:
            with self.access_lock:
                self._instrument.close()
        except visa.InvalidSession as e:
            logger.warning(
                "Could not close VISA connection, session invalid.", exc_info=e
            )

    @has_instrument
    def read_bytes(self) -> bytes:
        return cast("MessageBasedResource", self._instrument).read_raw()

    @has_instrument
    def write_bytes(self, data: bytes) -> int:
        return cast("MessageBasedResource", self._instrument).write_raw(data)

    @has_instrument
    def read(self) -> str:
        return cast("MessageBasedResource", self._instrument).read()

    @has_instrument
    def write(self, text: str) -> None:
        cast("MessageBasedResource", self._instrument).write(text)

    def read_all(
        self,
        _n_attempts_max: int | None = None,
        _attempt_interval_sec: Number | None = None,
    ) -> str | None:
        msg = "The 'read_all'-feature is not supported by visa."
        raise NotImplementedError(msg)

    @has_instrument
    def query(
        self,
        command: str,
        _n_attempts_max: int | None = None,
        _attempt_interval_sec: Number | None = None,
    ) -> str | None:
        with self.access_lock:
            return cast("MessageBasedResource", self._instrument).query(command)

    @has_instrument
    def write_multiple(self, *commands: str) -> None:
        """
        Write commands. No answer is read or expected.

        :param commands: one or more commands to send
        :raises VisaCommunicationError: when connection was not started
        """
        with self.access_lock:
            cast("MessageBasedResource", self._instrument).write(
                self._generate_cmd_string(commands)
            )

            # sleep small amount of time to not overload device
            sleep(self.WAIT_AFTER_WRITE)

    @has_instrument
    def query_multiple(self, *commands: str) -> str | tuple[str, ...]:
        """
        A combination of write(message) and read.

        :param commands: list of commands
        :return: list of values
        :raises VisaCommunicationError: when connection was not started, or when trying
            to issue too many commands at once.
        """
        cmd_string = self._generate_cmd_string(commands)
        with self.access_lock:
            return_string = cast("MessageBasedResource", self._instrument).query(
                cmd_string
            )

        if len(commands) == 1:
            return return_string

        return tuple(return_string.split(self.MULTI_COMMANDS_SEPARATOR))

    @classmethod
    def _generate_cmd_string(cls, command_list: tuple[str, ...]) -> str:
        """
        Generate the command string out of a tuple of strings.

        :param command_list: is the tuple containing multiple commands
        :return: the command string that can be sent via the protocol
        """

        if len(command_list) <= cls.MULTI_COMMANDS_MAX:
            return cls.MULTI_COMMANDS_SEPARATOR.join(command_list)

        msg = (
            f"Too many commands at once ({len(command_list)}). Max allowed: "
            f"{cls.MULTI_COMMANDS_MAX}."
        )
        raise VisaCommunicationError(msg)

    @has_instrument
    def spoll(self) -> int:
        """
        Execute serial poll on the device. Reads the status byte register STB. This
        is a fast function that can be executed periodically in a polling fashion.

        :return: integer representation of the status byte
        :raises VisaCommunicationError: when connection was not started
        """

        interface_type = self.config.interface_type

        if interface_type == VisaCommunicationConfig.InterfaceType.TCPIP_INSTR:
            with self.access_lock:
                return cast("MessageBasedResource", self._instrument).read_stb()

        if interface_type == VisaCommunicationConfig.InterfaceType.TCPIP_SOCKET:
            return int(self.query("*STB?"))

        msg = "Forgot to cover interface_type case?"
        raise VisaCommunicationError(msg)



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""Communication protocol for the KA3000P laboratory power supplies"""

import logging

from hvl_ccb.comm.base import SyncCommunicationProtocol
from hvl_ccb.comm.serial import SerialCommunication, SerialCommunicationConfig
from hvl_ccb.dev.ka3000p.base import KA3000PError

logger = logging.getLogger(__name__)


class KA3000PCommunicationConfig(SerialCommunicationConfig): ...


class KA3000PCommunication(SerialCommunication, SyncCommunicationProtocol):
    """Communication Protocol for KA3000P"""

    @staticmethod
    def config_cls():
        return KA3000PCommunicationConfig

    def query_not_none(
        self,
        command: str,
        n_attempts_max: int | None = None,
        attempt_interval_sec: float | None = None,
    ) -> str:
        value = self.query(command, n_attempts_max, attempt_interval_sec)

        if value is None:
            msg = (
                f'Did not recieved any responce on query "{command}", '
                "but expected something"
            )
            logger.error(msg)
            raise KA3000PError(msg)
        return value



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Communication and auxiliary classes for Technix
"""

import logging
from abc import ABC
from typing import NamedTuple, TypeAlias

from hvl_ccb import configdataclass
from hvl_ccb.comm import SyncCommunicationProtocol, SyncCommunicationProtocolConfig
from hvl_ccb.comm.serial import SerialCommunication, SerialCommunicationConfig
from hvl_ccb.comm.tcp import TcpCommunication, TcpCommunicationConfig
from hvl_ccb.dev import DeviceError
from hvl_ccb.utils.enum import ValueEnum
from hvl_ccb.utils.typing import Number

logger = logging.getLogger(__name__)


class TechnixError(DeviceError):
    """
    Technix related errors.
    """


class TechnixFaultError(TechnixError):
    """
    Raised when the fault flag was detected while the interlock is closed
    """


@configdataclass
class _TechnixCommunicationConfig(SyncCommunicationProtocolConfig):
    #: The terminator is CR
    terminator: bytes = b"\r"


class _TechnixCommunication(SyncCommunicationProtocol, ABC):
    """
    Generic communication class for Technix, which can be implemented via
    `TechnixSerialCommunication` or `TechnixTcpCommunication`
    """

    def query(
        self,
        command: str,
        n_attempts_max: int | None = None,
        attempt_interval_sec: Number | None = None,
    ) -> str:
        """
        Send a command to the interface and handle the status message.
        Possibly raises an error.

        :param command: Command to send
        :param n_attempts_max: Amount of attempts how often a non-empty text is tried to
            be read as answer
        :param attempt_interval_sec: time between the reading attempts
        :raises TechnixError: if the connection is broken
        :return: Answer from the interface
        """

        with self.access_lock:
            logger.debug(f"TechnixCommunication, send: '{command}'")
            answer: str | None = super().query(
                command,
                n_attempts_max=n_attempts_max,
                attempt_interval_sec=attempt_interval_sec,
            )  # string or None
            logger.debug(f"TechnixCommunication, receive: '{answer}'")
            if answer is None:
                msg = f"TechnixCommunication did get no answer on command: '{command}'"
                logger.error(msg)
                raise TechnixError(msg)
            return answer


@configdataclass
class TechnixSerialCommunicationConfig(
    _TechnixCommunicationConfig, SerialCommunicationConfig
):
    """
    Configuration for the serial communication for Technix
    """


class TechnixSerialCommunication(_TechnixCommunication, SerialCommunication):
    """
    Serial communication for Technix
    """

    @staticmethod
    def config_cls():
        return TechnixSerialCommunicationConfig


@configdataclass
class TechnixTcpCommunicationConfig(
    _TechnixCommunicationConfig, TcpCommunicationConfig
):
    """
    Configuration for the TCP communication for Technix
    """

    #: Port at which Technix is listening
    port: int = 4660


class TechnixTcpCommunication(TcpCommunication, _TechnixCommunication):
    """
    TCP communication for Technix
    """

    @staticmethod
    def config_cls():
        return TechnixTcpCommunicationConfig

    def open(self) -> None:
        super().open()

        while data := self.read_bytes():
            msg = f"Read telnet data at the beginning: {data!r}"
            logger.debug(msg)


_TechnixCommunicationClasses: TypeAlias = type[  # noqa: PYI047
    TechnixSerialCommunication | TechnixTcpCommunication
]


class _SetRegisters(ValueEnum):
    VOLTAGE = "d1"  # Output Voltage programming
    CURRENT = "d2"  # Output Current programming
    HVON = "P5"  # HV on
    HVOFF = "P6"  # HV off
    LOCAL = "P7"  # Local/remote mode
    INHIBIT = "P8"  # Inhibit


class _GetRegisters(ValueEnum):
    VOLTAGE = "a1"  # Output Voltage Monitor
    CURRENT = "a2"  # Output Current Monitor
    STATUS = "E"  # Image of the power supply logical status


class _Status(NamedTuple):
    """
    Container for the different statuses of the device. It can also handle the most
    recent reading of the voltage and current at the output.
    """

    inhibit: bool
    remote: bool
    hv_off: bool
    hv_on: bool
    output: bool
    open_interlock: bool
    fault: bool
    voltage_regulation: bool

    voltage: Number | None
    current: Number | None



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Classes for the BaseCube device.
"""

import logging
from collections import deque
from collections.abc import Sequence
from datetime import datetime, timedelta, timezone
from itertools import cycle, product
from time import sleep, time
from typing import Any, cast

from asyncua.sync import SyncNode

from hvl_ccb.comm.opc import (
    OpcUaCommunication,
    OpcUaCommunicationConfig,
    OpcUaSubHandler,
)
from hvl_ccb.configuration import configdataclass
from hvl_ccb.dev import SingleCommDevice
from hvl_ccb.utils.poller import Poller
from hvl_ccb.utils.typing import Number
from hvl_ccb.utils.validation import validate_bool, validate_number

from . import alarms, constants, earthing_stick
from .constants import (
    _CubeOpcEndpoint,
    _Door,
    _EarthingRod,
    _MeasurementChannel,
    _T13Socket,
)
from .errors import CubeRemoteControlError, CubeStatusChangeError, CubeStopError
from .support import _SupportPort

logger = logging.getLogger(__name__)


class _BaseCubeSubscriptionHandler(OpcUaSubHandler):
    """
    OPC Subscription handler for datachange events and normal events specifically
    implemented for the BaseCube devices.
    """

    def __init__(self) -> None:
        self.alarm_status = alarms._AlarmsOverview()

    def datachange_notification(self, node: SyncNode, val, data):
        """
        In addition to the standard operation (debug logging entry of the datachange),
        alarms are logged at INFO level using the alarm text.

        :param node: the node object that triggered the datachange event
        :param val: the new value
        :param data:
        """

        super().datachange_notification(node, val, data)

        # assume an alarm datachange
        try:
            alarm_number = alarms._Alarms(node.nodeid.Identifier).number
            setattr(
                self.alarm_status, f"alarm_{alarm_number}", alarms._AlarmStatus(val)
            )
            alarm_level = alarms._AlarmText.get_level(alarm_number)
            if val:
                alarm_text = alarms._AlarmText.get_coming_message(alarm_number)
            else:
                alarm_text = alarms._AlarmText.get_going_message(alarm_number)

            alarm_log_postfix = ""
            if logger.root.level <= logging.DEBUG:
                alarm_log_postfix = f" (opc alarm {alarm_number})"  # pragma: no cover

            if alarm_level == logging.DEBUG:
                logger.debug(f"{alarm_text}{alarm_log_postfix}")
            elif alarm_level == logging.INFO:
                logger.info(f"{alarm_text}{alarm_log_postfix}")
            elif alarm_level >= logging.WARNING:
                logger.warning(f"{alarm_text}{alarm_log_postfix}")
        except ValueError:
            # not any of Alarms node IDs
            pass
        else:
            return

        id_ = node.nodeid.Identifier

        # assume a status datachange
        if id_ == str(constants._Safety.STATUS):
            new_status = constants.SafetyStatus(val)
            logger.info(f"Safety: {new_status.name}")
            return

        # assume an earthing stick status datachange
        for i in earthing_stick._EarthingStick._STICKS:
            if id_ in earthing_stick._EarthingStick("", i)._CMD_STATUS:
                new_status = earthing_stick.SwitchStatus(val)
                logger.info(f"Earthing Stick {i}: {new_status.name}")
                return


@configdataclass
class BaseCubeConfiguration:
    """
    Configuration dataclass for the BaseCube devices.
    """

    #: Namespace of the OPC variables, typically this is 3 (coming from Siemens)
    namespace_index: int = 3

    polling_delay_sec: Number = 5.0
    polling_interval_sec: Number = 1.0

    timeout_status_change: Number = 6
    timeout_interval: Number = 0.1

    noise_level_measurement_channel_1: Number = 100
    noise_level_measurement_channel_2: Number = 100
    noise_level_measurement_channel_3: Number = 100
    noise_level_measurement_channel_4: Number = 100

    def clean_values(self) -> None:
        if self.namespace_index < 0:
            msg = "Index of the OPC variables namespace needs to be a positive integer."
            raise ValueError(msg)
        if self.polling_interval_sec <= 0:
            msg = "Polling interval needs to be positive."
            raise ValueError(msg)
        if self.polling_delay_sec < 0:
            msg = "Polling delay needs to be not negative."
            raise ValueError(msg)
        if self.timeout_interval <= 0:
            msg = "Timeout interval for status change needs to be positive."
            raise ValueError(msg)
        if self.timeout_status_change < 0:
            msg = "Timeout for status change needs to be not negative."
            raise ValueError(msg)
        validate_number(
            "Noise Level Measurement Channel 1",
            self.noise_level_measurement_channel_1,
            logger=logger,
        )
        validate_number(
            "Noise Level Measurement Channel 2",
            self.noise_level_measurement_channel_2,
            logger=logger,
        )
        validate_number(
            "Noise Level Measurement Channel 3",
            self.noise_level_measurement_channel_3,
            logger=logger,
        )
        validate_number(
            "Noise Level Measurement Channel 4",
            self.noise_level_measurement_channel_4,
            logger=logger,
        )


@configdataclass
class BaseCubeOpcUaCommunicationConfig(OpcUaCommunicationConfig):
    """
    Communication protocol configuration for OPC UA, specifications for the BaseCube
    devices.
    """

    #: Subscription handler for data change events
    sub_handler: OpcUaSubHandler = _BaseCubeSubscriptionHandler()
    endpoint_name: _CubeOpcEndpoint = _CubeOpcEndpoint.BASE_CUBE  # type: ignore[assignment]


class BaseCubeOpcUaCommunication(OpcUaCommunication):
    """
    Communication protocol specification for BaseCube devices.
    """

    @staticmethod
    def config_cls():
        return BaseCubeOpcUaCommunicationConfig


class BaseCube(SingleCommDevice):
    """
    Base class for Cube variants.
    """

    OPC_MIN_YEAR = 1990
    OPC_MAX_YEAR = 2089

    def __init__(self, com, dev_config=None) -> None:
        """
        Constructor for BaseCube.

        :param com: the communication protocol or its configuration
        :param dev_config: the device configuration
        """

        super().__init__(com, dev_config)

        self._status_poller = Poller(
            self._spoll_handler,
            polling_delay_sec=self.config.polling_delay_sec,
            polling_interval_sec=self.config.polling_interval_sec,
        )
        self._toggle = cycle([False, True])
        self._message_len = len(constants.MessageBoard)
        self._status_board = [""] * self._message_len
        self._message_board = deque([""] * self._message_len, maxlen=self._message_len)

        # create earthing sticks
        self.earthing_stick_1 = earthing_stick._EarthingStick(self, 1)
        self.earthing_stick_2 = earthing_stick._EarthingStick(self, 2)
        self.earthing_stick_3 = earthing_stick._EarthingStick(self, 3)
        self.earthing_stick_4 = earthing_stick._EarthingStick(self, 4)
        self.earthing_stick_5 = earthing_stick._EarthingStick(self, 5)
        self.earthing_stick_6 = earthing_stick._EarthingStick(self, 6)

        # create support ports, each port has two inputs and outputs
        self.support_1 = _SupportPort(self, 1)
        self.support_2 = _SupportPort(self, 2)
        self.support_3 = _SupportPort(self, 3)
        self.support_4 = _SupportPort(self, 4)
        self.support_5 = _SupportPort(self, 5)
        self.support_6 = _SupportPort(self, 6)

        # create measurement channels
        self.measurement_ch_1 = _MeasurementChannel(
            self, 1, input_noise=self.config.noise_level_measurement_channel_1
        )
        self.measurement_ch_2 = _MeasurementChannel(
            self, 2, input_noise=self.config.noise_level_measurement_channel_2
        )
        self.measurement_ch_3 = _MeasurementChannel(
            self, 3, input_noise=self.config.noise_level_measurement_channel_3
        )
        self.measurement_ch_4 = _MeasurementChannel(
            self, 4, input_noise=self.config.noise_level_measurement_channel_4
        )

    @staticmethod
    def default_com_cls():
        return BaseCubeOpcUaCommunication

    @staticmethod
    def config_cls():
        return BaseCubeConfiguration

    def start(self) -> None:
        """
        Starts the device. Sets the root node for all OPC read and write commands to
        the Siemens PLC object node which holds all our relevant objects and variables.
        """

        logger.info("Starting Cube")
        super().start()

        logger.debug("Add monitoring nodes")
        self.com.init_monitored_nodes(constants._CEE16, self.config.namespace_index)
        # add T13 sockets
        for socket in _T13Socket._SOCKETS:
            self.com.init_monitored_nodes(
                _T13Socket(socket)._CMD, self.config.namespace_index
            )
        for io, port, contact in product(
            _SupportPort._IOS, _SupportPort._PORTS, _SupportPort._CONTACTS
        ):
            self.com.init_monitored_nodes(
                _SupportPort("", port)._cmd(io, contact), self.config.namespace_index
            )
        self.com.init_monitored_nodes(
            map(str, constants._Safety),
            self.config.namespace_index,
        )
        self.com.init_monitored_nodes(
            str(constants._Errors.MESSAGE), self.config.namespace_index
        )
        self.com.init_monitored_nodes(
            str(constants._Errors.WARNING), self.config.namespace_index
        )
        self.com.init_monitored_nodes(
            str(constants._Errors.STOP), self.config.namespace_index
        )
        self.com.init_monitored_nodes(
            map(str, alarms._Alarms), self.config.namespace_index
        )
        for i in earthing_stick._EarthingStick._STICKS:
            self.com.init_monitored_nodes(
                earthing_stick._EarthingStick("", i)._CMD_STATUS,
                self.config.namespace_index,
            )

        self._set_remote_control(True)
        logger.info("Finished starting")

        self._set_current_time()
        logger.debug("Sent system time to Cube.")

    def stop(self) -> None:
        """
        Stop the Cube device. Deactivates the remote control and closes the
        communication protocol.

        :raises CubeStopError: when the cube is not in the correct status
            to stop the operation
        """
        status = self._status
        if status not in constants.STOP_SAFETY_STATUSES:
            msg = (
                "Cube needs to be in status "
                f"{' or '.join(s.name for s in constants.STOP_SAFETY_STATUSES)} "
                f"to close the connection to the device, but is in status {status.name}"
            )
            logger.error(msg)
            raise CubeStopError(msg)
        try:
            self._set_remote_control(False)
        finally:
            super().stop()
            logger.info("Stopping Cube")

    def _spoll_handler(self) -> None:
        """
        Cube poller handler; change one byte on a Cube.
        """
        self.write(constants._OpcControl.LIVE, next(self._toggle))

    def read(self, node_id: str) -> Any:
        """
        Local wrapper for the OPC UA communication protocol read method.

        :param node_id: the id of the node to read.
        :return: the value of the variable
        """

        logger.debug(f"Read from node ID {node_id} ...")
        result = self.com.read(str(node_id), self.config.namespace_index)
        logger.debug(f"Read from node ID {node_id}: {result}")
        return result

    def write(self, node_id, value) -> None:
        """
        Local wrapper for the OPC UA communication protocol write method.

        :param node_id: the id of the node to write
        :param value: the value to write to the variable
        """
        logger.debug(f"Write to node ID {node_id}: {value}")
        self.com.write(str(node_id), self.config.namespace_index, value)

    @classmethod
    def datetime_to_opc(cls, time_dt: datetime) -> list[int]:
        """
        Converts python datetime format into opc format (list of 8 integers) as defined
        in the following link:
        https://support.industry.siemens.com/cs/mdm/109798671?c=133950752267&lc=de-WW
        Each byte corresponds to one list entry.
        [yy, MM, dd, hh, mm, ss, milliseconds, weekday]
        Milliseconds and Weekday are not used, as this precision / information is not
        needed.
        The conversion of the numbers is special. Each decimal number is treated as it
        would be a hex-number and then converted back to decimal.
        This is tested with the used PLC in the BaseCube.
        yy: 0 to 99 (0 -> 2000, 89 -> 2089, 90 -> 1990, 99 -> 1999)
        MM: 1 to 12
        dd: 1 to 31
        hh: 0 to 23
        mm: 0 to 59
        ss: 0 to 59

        :param time_dt: time to be converted
        :return: time in opc list format
        """

        validate_number("year", time_dt.year, (cls.OPC_MIN_YEAR, cls.OPC_MAX_YEAR), int)

        time_tuple = (
            int(str(time_dt.year)[2:]),
            time_dt.month,
            time_dt.day,
            time_dt.hour,
            time_dt.minute,
            time_dt.second,
            0,
            0,
        )
        return [int(f"0x{time_comp}", base=0) for time_comp in time_tuple]

    def _set_current_time(self) -> None:
        """
        Send current UTC time of host computer to Cube. As the time is only
        synchronized during the startup and after the polling started, the polling delay
        is added to the current time
        """
        time_opc = self.datetime_to_opc(
            datetime.now(timezone.utc)
            + timedelta(seconds=self.config.polling_delay_sec)
        )
        self.write(constants._OpcControl.TIME, time_opc)

    def _set_remote_control(self, state: bool) -> None:
        """
        Enable or disable remote control for the Cube. This will effectively
        display a message on the touchscreen HMI.

        :param state: desired remote control state
        :raises TypeError: when state is not of type bool
        :raises CubeRemoteControlError: when the remote control
            cannot be (de-)activated
        """
        validate_bool("state", state, logger)
        if not state and self._status not in constants.STOP_SAFETY_STATUSES:
            status = self._status
            msg = (
                "Cube needs to be in status 'GREEN_NOT_READY' or 'GREEN_READY' "
                f"to turn off the remote control, but is in '{status.name}'"
            )
            logger.error(msg)
            raise CubeRemoteControlError(msg)
        can_write = False
        try:
            self.write(constants._OpcControl.ACTIVE, state)
            can_write = True
        finally:
            if state:
                if not can_write:
                    msg = "Remote control cannot be enabled"  # pragma: no cover
                    logger.error(msg)  # pragma: no cover
                    raise CubeRemoteControlError(msg)  # pragma: no cover
                was_not_polling = self._status_poller.start_polling()
                if not was_not_polling:
                    msg = "Remote control already enabled"  # pragma: no cover
                    logger.error(msg)  # pragma: no cover
                    raise CubeRemoteControlError(msg)  # pragma: no cover
            else:
                was_polling = self._status_poller.stop_polling()
                self._status_poller.wait_for_polling_result()
                if not was_polling:
                    msg = "Remote control already disabled"  # pragma: no cover
                    logger.error(msg)  # pragma: no cover
                    raise CubeRemoteControlError(msg)  # pragma: no cover

    # creates the T13 power sockets
    t13_socket_1 = _T13Socket(socket=1)
    t13_socket_2 = _T13Socket(socket=2)
    t13_socket_3 = _T13Socket(socket=3)

    @property
    def cee16_socket(self) -> bool:
        """
        Read the on-state of the IEC CEE16 three-phase power socket.

        :return: the on-state of the CEE16 power socket
        """

        value = bool(self.read(constants._CEE16))
        state = "ON" if value else "OFF"
        logger.info(f"CEE 16 A Power Socket is {state}")
        return value

    @cee16_socket.setter
    def cee16_socket(self, value) -> None:
        """
        Switch the IEC CEE16 three-phase power socket on or off.

        :param value: desired on-state of the power socket
        :raises TypeError: if state is not of type bool
        """

        validate_bool("state", value, logger)

        self.write(constants._CEE16, value)
        state_str = "ON" if value else "OFF"
        logger.info(f"CEE 16 A Power Socket is switched {state_str}")

    @property
    def _status(self) -> constants.SafetyStatus:
        """
        Get the safety circuit status of the Cube.
        For internal use only, without logging

        :return: the safety status of the Cube's state machine.
        """
        return constants.SafetyStatus(self.read(constants._Safety.STATUS))

    @property
    def status(self) -> constants.SafetyStatus:
        """
        Get the safety circuit status of the Cube.
        This methods is for the user.

        :return: the safety status of the Cube's state machine.
        """

        value = self._status
        logger.info(f"Safety Status is {value.name}")
        return value

    def _switch_safety_status(
        self,
        switch: constants._SafetyStatusTransition,
        raise_state: bool = True,
    ):
        """
        Internal method to switch the safety status of the cube.
        Checks are performed if the switch can be executed.

        :param switch: indicates the transition
        :param raise_state: True will raise the state, False will lower the state
        :raise CubeStatusChangeError: if the status change was not successful or
            the status cannot be changed because the cube is in the wrong safety status
            for the queried operation
        """
        source = constants._SafetyStatusTransition(switch).source
        target = constants._SafetyStatusTransition(switch).target
        actual_status = self._status
        if raise_state and actual_status is not source:
            msg = (
                f"Cube needs to be in status '{source.name}' in order to switch "
                f"to '{target.name}', but is in '{actual_status.name}'"
            )
            logger.error(msg)
            raise CubeStatusChangeError(msg)
        if not raise_state and actual_status is not target:
            msg = (
                f"Cube needs to be in status '{target.name}' in order to switch "
                f"to '{source.name}', but is in '{actual_status.name}'"
            )
            logger.error(msg)
            raise CubeStatusChangeError(msg)

        status = target if raise_state else source
        logger.info(f"Status of Cube will be changed to {status.name}")
        self.write(switch.command, raise_state)

        start_time = time()
        while (
            self._status is not status
            and time() - start_time < self.config.timeout_status_change
        ):
            sleep(self.config.timeout_interval)
        actual_status = self._status
        if actual_status is status:
            logger.info(f"Status successfully changed to {status.name}.")
        else:
            msg = (
                f"Tried to change status to {status.name}, "
                f"but Cube status is {actual_status.name}"
            )
            logger.error(msg)
            raise CubeStatusChangeError(msg)

    @property
    def ready(self) -> bool | None:
        """
        Indicates if 'ready' is activated. 'ready' means locket safety circuit,
        red lamps, but high voltage still off.

        :return: `True` if ready is activated (RED_READY),
            `False` if ready is deactivated (GREEN_READY),
            `None` otherwise
        """
        status = self.status
        if status == constants.SafetyStatus.RED_READY:
            return True
        if status == constants.SafetyStatus.GREEN_READY:
            return False
        return None

    @ready.setter
    def ready(self, state: bool) -> None:
        """
        Set ready state. Ready means locket safety circuit, red lamps, but high voltage
        still off.

        :param state: set ready state
        :raises CubeStatusChangeError: if `state=True` and cube is not in GREEN_READY or
            if `state=False` and cube is not in RED_READY
        """
        validate_bool("state", state, logger)
        self._switch_safety_status(
            switch=cast(
                "constants._SafetyStatusTransition",
                constants._SafetyStatusTransition.SWITCH_TO_READY,
            ),
            raise_state=state,
        )

    @property
    def operate(self) -> bool | None:
        """
        Indicates if 'operate' is activated. 'operate' means locket safety circuit,
        red lamps, high voltage on and locked safety switches.

        :return: `True` if operate is activated (RED_OPERATE),
            `False` if ready is deactivated (RED_READY),
            `None` otherwise
        """
        status = self.status
        if status == constants.SafetyStatus.RED_OPERATE:
            return True
        if status == constants.SafetyStatus.RED_READY:
            return False
        return None

    @operate.setter
    def operate(self, state: bool) -> None:
        """
        Set operate state. This will turn on the high
        voltage and close the safety switches.

        :param state: set operate state
        :raises CubeStatusChangeError: if `state=True` and cube is not in RED_READY or
            if `state=False` and cube is not in RED_OPERATE
        """
        validate_bool("state", state, logger)
        self._switch_safety_status(
            switch=cast(
                "constants._SafetyStatusTransition",
                constants._SafetyStatusTransition.SWITCH_TO_OPERATE,
            ),
            raise_state=state,
        )

    @property
    def breakdown_detection_active(self) -> bool:
        """
        Get the state of the breakdown detection functionality. Returns True if it is
        enabled, False otherwise.

        :return: state of the breakdown detection functionality
        """

        value = self.read(constants._BreakdownDetection.ACTIVATED)
        status = "" if value else "NOT "
        logger.info(f"Breakdown Detection Unit is {status}activated")
        return value

    @property
    def breakdown_detection_triggered(self) -> bool:
        """
        See if breakdown detection unit has been triggered. Returns True if it is
        triggered, False otherwise.

        :return: trigger status of the breakdown detection unit
        """
        value = self.read(constants._BreakdownDetection.TRIGGERED)
        status = "" if value else "NOT "
        logger.info(f"Breakdown Detection Unit is {status}triggered")
        return value

    def breakdown_detection_reset(self) -> None:
        """
        Reset the breakdown detection circuitry so that it is ready
        to detect breakdowns again.
        """

        self.write(constants._BreakdownDetection.RESET, True)
        sleep(0.1)
        self.write(constants._BreakdownDetection.RESET, False)
        logger.info("The Breakdown Detection Unit is reset")

    def quit_error(self) -> None:
        """
        Quits errors that are active on the Cube.
        """

        logger.info("Quit Errors of Cube")
        self.write(constants._Errors.QUIT, True)
        sleep(0.1)
        self.write(constants._Errors.QUIT, False)

    # create doors
    door_1_status = _Door(1, "door status")
    door_2_status = _Door(2, "door status")
    door_3_status = _Door(3, "door status")

    # create earthing rods
    earthing_rod_1_status = _EarthingRod(1, "earthing rod status")
    earthing_rod_2_status = _EarthingRod(2, "earthing rod status")
    earthing_rod_3_status = _EarthingRod(3, "earthing rod status")

    def set_status_board(
        self,
        msgs: list[str],
        pos: list[int] | None = None,
        clear_board: bool = True,
        display_board: bool = True,
    ) -> None:
        """
        Sets and displays a status board. The messages and the position of the message
        can be defined.

        :param msgs: list of strings
        :param pos: list of integers [0...14]
        :param clear_board: clear unspecified lines if `True` (default), keep otherwise
        :param display_board: display new status board if `True` (default)
        :raises ValueError: if there are too many messages or the positions indices are
            invalid.
        """
        # validate inputs
        if len(msgs) > self._message_len:
            msg = (
                f"Too many message: {len(msgs)} given, "
                f"max. {self._message_len} allowed."
            )
            raise ValueError(msg)
        if pos and not all(0 < p < self._message_len for p in pos):
            msg = f"Messages positions out of 0...{self._message_len} range"
            raise ValueError(msg)

        if clear_board:
            self._status_board = [""] * self._message_len

        # update status board
        if not pos:
            pos = list(range(len(msgs)))
        for num, msg in zip(pos, msgs, strict=False):
            self._status_board[num] = msg
        if display_board:
            self.display_status_board()

    def display_status_board(self) -> None:
        """
        Display status board.
        """

        self._display_messages(self._status_board)
        logger.info("Cube HMI is now displaying the Status Board")

    def set_message_board(self, msgs: list[str], display_board: bool = True) -> None:
        """
        Fills messages into message board that display that 15 newest messages with
        a timestamp.

        :param msgs: list of strings
        :param display_board: display 15 newest messages if `True` (default)
        :raises ValueError: if there are too many messages or the positions indices are
            invalid.
        """
        # validate inputs
        if len(msgs) > self._message_len:
            msg = (
                f"Too many message: {len(msgs)} given, "
                f"max. {self._message_len} allowed."
            )
            raise ValueError(msg)

        timestamp = datetime.now().time().strftime("%H:%M:%S")
        # append messages in the same order as given, not reversed
        self._message_board.extendleft(f"{timestamp}: {msg}" for msg in reversed(msgs))

        if display_board:
            self.display_message_board()

    def display_message_board(self) -> None:
        """
        Display 15 newest messages
        """

        self._display_messages(self._message_board)
        logger.info("Cube HMI is now displaying the Message Board")

    def _display_messages(self, messages: Sequence[str]) -> None:
        """
        Display given messages on message board

        :param messages: sequence of messages to display
        """
        # Note: cannot zip(constants.MessageBoard, messages) as enum instances are
        #       sorted by name, hence after after `line_1` comes `line_10`, not `line_2`
        for n, msg in enumerate(messages):
            line = constants.MessageBoard.line(n + 1)
            self.write(line, msg)

    def active_alarms(self, human_readable: bool = True) -> list[int | str]:
        """
        Displays all active alarms / messages.

        :param human_readable: `True` for human readable message,
            `False` for corresponding integer
        :return: list with active alarms
        """
        validate_bool("human_readable", human_readable, logger=logger)
        active_alarms = []
        for i in alarms._Alarms.range():
            if getattr(self._com.config.sub_handler.alarm_status, f"alarm_{i}"):
                if human_readable:
                    i = alarms._AlarmText.get_coming_message(i)
                active_alarms.append(i)
        return active_alarms



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
""" """

import logging
import time
from collections.abc import Generator
from typing import cast

import libtiepie as ltp
import numpy as np
import numpy.typing as npt
from aenum import IntEnum
from libtiepie import oscilloscope as ltp_osc

from hvl_ccb.comm import NullCommunicationProtocol
from hvl_ccb.dev import SingleCommDevice
from hvl_ccb.utils.enum import NameEnum
from hvl_ccb.utils.typing import Number
from hvl_ccb.utils.validation import validate_number

from .base import (
    TiePieDeviceConfig,
    TiePieDeviceType,
    TiePieError,
    _LtpDeviceReturnType,
    _require_dev_handle,
    _verify_via_libtiepie,
    get_device_by_serial_number,
    wrap_libtiepie_exception,
)
from .channel import TiePieOscilloscopeChannelConfig
from .utils import PublicPropertiesReprMixin

logger = logging.getLogger(__name__)


class TiePieOscilloscopeResolution(IntEnum):
    EIGHT_BIT = 8
    TWELVE_BIT = 12
    FOURTEEN_BIT = 14
    SIXTEEN_BIT = 16


class TiePieOscilloscopeAutoResolutionModes(NameEnum, init="value description"):  # type: ignore[call-arg]
    UNKNOWN = ltp.ARM_UNKNOWN, "Unknown"
    DISABLED = ltp.ARM_DISABLED, "Disabled"
    NATIVEONLY = ltp.ARM_NATIVEONLY, "Native only"
    ALL = ltp.ARM_ALL, "All"


class TiePieOscilloscopeConfigLimits:
    """
    Default limits for oscilloscope parameters.
    """

    def __init__(self, dev_osc: ltp_osc.Oscilloscope) -> None:
        self.record_length = (0, dev_osc.record_length_max)
        self.sample_rate = (0, dev_osc.sample_rate_max)  # [samples/s]
        self.pre_sample_ratio = (0, 1)
        # self.trigger_delay = (0, dev_osc.trigger_delay_max)  # trigger_delay is not
        # available for all instruments, cf. API from TiePie:
        # Functions » Oscilloscope » Trigger » Delay


class TiePieOscilloscopeConfig(PublicPropertiesReprMixin):
    """
    Oscilloscope's configuration with cleaning of values in properties setters.
    """

    def __init__(self, dev_osc: ltp_osc.Oscilloscope) -> None:
        self.dev_osc: ltp_osc.Oscilloscope = dev_osc
        self.param_lim: TiePieOscilloscopeConfigLimits = TiePieOscilloscopeConfigLimits(
            dev_osc=dev_osc
        )

    def clean_pre_sample_ratio(self, pre_sample_ratio: float) -> float:
        validate_number(
            "pre sample ratio",
            pre_sample_ratio,
            self.param_lim.pre_sample_ratio,
            logger=logger,
        )
        return float(pre_sample_ratio)

    @property
    def pre_sample_ratio(self) -> float:
        return self.dev_osc.pre_sample_ratio

    @pre_sample_ratio.setter
    def pre_sample_ratio(self, pre_sample_ratio: float) -> None:
        """
        Set pre sample ratio

        :param pre_sample_ratio: pre sample ratio numeric value.
        :raise ValueError: If `pre_sample_ratio` is not a number between 0 and 1
            (inclusive).
        """
        self.dev_osc.pre_sample_ratio = self.clean_pre_sample_ratio(pre_sample_ratio)
        logger.info(f"Pre-sample ratio is set to {pre_sample_ratio}.")

    def clean_record_length(self, record_length: Number) -> int:
        validate_number(
            "record length",
            record_length,
            limits=self.param_lim.record_length,
            logger=logger,
        )

        if not (float(record_length).is_integer()):
            msg = (
                "The record_length has to be a value, that can be cast "
                "into an integer without significant precision loss; "
                f"but {record_length:_d} was assigned."
            )
            raise ValueError(msg)

        return cast(
            "int",
            _verify_via_libtiepie(self.dev_osc, "record_length", int(record_length)),
        )

    @property
    def record_length(self) -> int:
        return self.dev_osc.record_length

    @record_length.setter
    def record_length(self, record_length: int) -> None:
        record_length = self.clean_record_length(record_length)
        self.dev_osc.record_length = record_length
        logger.info(f"Record length is set to {record_length:_d} Sa.")

    @staticmethod
    def clean_resolution(
        resolution: int | TiePieOscilloscopeResolution,
    ) -> TiePieOscilloscopeResolution:
        if not isinstance(resolution, TiePieOscilloscopeResolution):
            validate_number("resolution", resolution, number_type=int, logger=logger)
        return TiePieOscilloscopeResolution(resolution)

    @property
    def resolution(self) -> TiePieOscilloscopeResolution:
        return self.dev_osc.resolution

    @resolution.setter
    def resolution(self, resolution: int | TiePieOscilloscopeResolution) -> None:
        """
        Setter for resolution of the Oscilloscope.

        :param resolution: resolution integer.
        :raises ValueError: if resolution is not one of
            `TiePieOscilloscopeResolution` instance or integer values
        """
        self.dev_osc.resolution = self.clean_resolution(resolution)
        logger.info(f"Resolution is set to {self.dev_osc.resolution} bit.")

    @staticmethod
    def clean_auto_resolution_mode(
        auto_resolution_mode: int | TiePieOscilloscopeAutoResolutionModes,
    ) -> TiePieOscilloscopeAutoResolutionModes:
        if not isinstance(auto_resolution_mode, TiePieOscilloscopeAutoResolutionModes):
            validate_number(
                "auto resolution mode",
                auto_resolution_mode,
                number_type=int,
                logger=logger,
            )
        if isinstance(auto_resolution_mode, bool):
            msg = "Auto resolution mode cannot be of boolean type"
            logger.error(msg)
            raise TypeError
        return TiePieOscilloscopeAutoResolutionModes(auto_resolution_mode)

    @property
    def auto_resolution_mode(self) -> TiePieOscilloscopeAutoResolutionModes:
        return TiePieOscilloscopeAutoResolutionModes(self.dev_osc.auto_resolution_mode)

    @auto_resolution_mode.setter
    def auto_resolution_mode(self, auto_resolution_mode):
        self.dev_osc.auto_resolution_mode = self.clean_auto_resolution_mode(
            auto_resolution_mode
        ).value
        logger.info(f"Auto resolution mode is set to {auto_resolution_mode}.")

    def clean_sample_rate(self, sample_rate: float) -> float:
        validate_number(
            "sample rate",
            sample_rate,
            self.param_lim.sample_rate,
            logger=logger,
        )
        sample_rate = _verify_via_libtiepie(self.dev_osc, "sample_rate", sample_rate)
        return float(sample_rate)

    @property
    def sample_rate(self) -> float:
        return self.dev_osc.sample_rate

    @sample_rate.setter
    def sample_rate(self, sample_rate: float):
        """
        Set sample rate of the oscilloscope.

        :param sample_rate: rate to set
        :raises ValueError: when rate is not in device range
        """
        sample_rate = self.clean_sample_rate(sample_rate)
        self.dev_osc.sample_rate = sample_rate
        logger.info(f"Sample rate is set to {sample_rate:_.3f} Sa/s.")

    @property
    def sample_frequency(self) -> float:
        """For backwards compatibility. Use `sample_rate` instead"""
        logger.warning(
            "The usage of `sample_frequency` is deprecated, use "
            "`sample_rate` instead. In future versions this will raise an "
            "`AttributeError`."
        )
        return self.sample_rate

    @sample_frequency.setter
    def sample_frequency(self, sample_frequency: float):
        """For backwards compatibility. Use `sample_rate` instead"""
        logger.warning(
            "The usage of `sample_frequency` is deprecated, use "
            "`sample_rate` instead. In future versions this will raise an "
            "`AttributeError`."
        )
        self.sample_rate = sample_frequency

    def clean_trigger_timeout(self, trigger_timeout: Number | None) -> float:
        if trigger_timeout in (None, ltp.const.TO_INFINITY):
            # infinite timeout: `TO_INFINITY = -1` in `libtiepie.const`
            trigger_timeout = ltp.const.TO_INFINITY
        else:
            validate_number(
                "trigger timeout",
                trigger_timeout,
                limits=(0, None),
                logger=logger,
            )
        trigger_timeout = _verify_via_libtiepie(
            self.dev_osc.trigger, "timeout", cast("Number", trigger_timeout)
        )
        return float(trigger_timeout)

    @property
    def trigger_timeout(self) -> float | None:
        if self.dev_osc.trigger.timeout == ltp.const.TO_INFINITY:
            return None
        return self.dev_osc.trigger.timeout

    @trigger_timeout.setter
    def trigger_timeout(self, trigger_timeout: Number | None) -> None:
        """
        Set trigger time-out.

        :param trigger_timeout: Trigger timeout value, in seconds; `0`  forces
            trigger to start immediately after starting a measurement;
            None leads to no timeout
        :raise ValueError: If trigger timeout is not a non-negative real number.
        """
        trigger_timeout = self.clean_trigger_timeout(trigger_timeout)
        self.dev_osc.trigger.timeout = trigger_timeout
        if trigger_timeout == ltp.const.TO_INFINITY:
            logger.info("Trigger timeout is set to \u221e (INFINITY) s.")
        else:
            logger.info(f"Trigger timeout is set to {trigger_timeout} s.")


class TiePieOscilloscope(SingleCommDevice):
    """
    TiePie oscilloscope.

    A wrapper for TiePie oscilloscopes, based on the class
    `libtiepie.oscilloscope.Oscilloscope` with simplifications for starting of the
    device (using serial number) and managing mutable configuration of both the
    device and its channels, including extra validation and typing hints support for
    configurations.

    Note that, in contrast to `libtiepie` library, since all physical TiePie devices
    include an oscilloscope, this is the base class for all physical TiePie devices.
    The additional TiePie sub-devices: "Generator" is mixed-in to this base class in
    subclasses.

    The channels use `1..N` numbering (not `0..N-1`), as in, e.g., the Multi Channel
    software.
    """

    @staticmethod
    def config_cls() -> type[TiePieDeviceConfig]:
        return TiePieDeviceConfig

    @staticmethod
    def default_com_cls() -> type[NullCommunicationProtocol]:
        return NullCommunicationProtocol

    def __init__(self, com, dev_config) -> None:
        """
        Constructor for a TiePie device.
        """
        super().__init__(com, dev_config)

        self._osc: ltp_osc.Oscilloscope | None = None

        self.config_osc: TiePieOscilloscopeConfig | None = None
        """
        Oscilloscope's dynamical configuration.
        """

        self.config_osc_channel_dict: dict[int, TiePieOscilloscopeChannelConfig] = {}
        """
        Channel configuration.
        A `dict` mapping actual channel number, numbered `1..N`, to channel
        configuration. The channel info is dynamically read from the device only on
        the first `start()`; beforehand the `dict` is empty.
        """

    @_require_dev_handle(TiePieDeviceType.OSCILLOSCOPE)
    def _osc_config_setup(self) -> None:
        """
        Setup dynamical configuration for the connected oscilloscope.
        """
        self.config_osc = TiePieOscilloscopeConfig(
            cast("ltp_osc.Oscilloscope", self._osc)
        )
        for n in range(1, self.n_channels + 1):
            self.config_osc_channel_dict[n] = TiePieOscilloscopeChannelConfig(
                ch_number=n,
                channel=cast("ltp_osc.Oscilloscope", self._osc).channels[n - 1],
            )

    def _osc_config_teardown(self) -> None:
        """
        Teardown dynamical configuration for the oscilloscope.
        """
        self.config_osc = None
        self.config_osc_channel_dict = {}

    def _osc_close(self) -> None:
        """
        Close the wrapped `libtiepie` oscilloscope.
        """
        if self._osc is not None:
            del self._osc
            self._osc = None

    def _get_device_by_serial_number(
        self,
        # Note: TiePieDeviceType aenum as a tuple to define a return value type
        ltp_device_type: tuple[int, _LtpDeviceReturnType],
    ) -> _LtpDeviceReturnType:
        """
        Wrapper around `get_device_by_serial_number` using this device's config options.

        :return: A `libtiepie` device object specific to a class it is called on.
        """
        return get_device_by_serial_number(
            self.config.serial_number,
            ltp_device_type,
            n_max_try_get_device=self.config.n_max_try_get_device,
            wait_sec_retry_get_device=self.config.wait_sec_retry_get_device,
        )

    @wrap_libtiepie_exception
    def start(self) -> None:
        """
        Start the oscilloscope.
        """
        logger.info(f"Starting {self}")
        super().start()
        logger.info(
            f"Starting oscilloscope with serial number {self.config.serial_number}"
        )

        self._osc = self._get_device_by_serial_number(TiePieDeviceType.OSCILLOSCOPE)

        # Check for block measurement support if required
        if self.config.require_block_measurement_support and not (
            self._osc.measure_modes & ltp.MM_BLOCK
        ):
            self._osc_close()
            msg = (
                f"Oscilloscope with serial number {self.config.serial_number} does not "
                "have required block measurement support."
            )
            logger.error(msg)
            raise TiePieError(msg)

        self._osc_config_setup()

    @wrap_libtiepie_exception
    def stop(self) -> None:
        """
        Stop the oscilloscope.
        """
        logger.info(f"Stopping {self}")
        logger.info("Stopping oscilloscope")

        self._osc_config_teardown()
        self._osc_close()

        super().stop()

    @staticmethod
    @wrap_libtiepie_exception
    def list_devices() -> ltp.devicelist.DeviceList:
        """
        List available TiePie devices.

        :return: libtiepie up to date list of devices
        """
        ltp.network.auto_detect_enabled = True
        device_list = ltp.device_list
        device_list.update()

        # log devices list
        if device_list:
            logger.info("Available devices:\n")

            for item in ltp.device_list:
                logger.info(f"  Name:              {item.name}")
                logger.info(f"  Serial number:     {item.serial_number}")
                logger.info(f"  Available types:   {ltp.device_type_str(item.types)}")
                if item.has_server:
                    logger.info(
                        f"  Server:            {item.server.url}({item.server.name})"
                    )
                logger.info(
                    "  Can be opened as "
                    f"Oscilloscope: {item.can_open(ltp.DEVICETYPE_OSCILLOSCOPE)}\n"
                )

        else:
            logger.info("No devices found!")

        return device_list

    @wrap_libtiepie_exception
    @_require_dev_handle(TiePieDeviceType.OSCILLOSCOPE)
    def start_measurement(self) -> None:
        """
        Start a measurement using set configuration.

        :raises TiePieError: when device is not started, when measurement is already
        running, or when status of underlying device gives an error.
        """
        if self.is_measurement_running():
            msg = "TiePie measurement is already running"
            raise TiePieError(msg)
        cast("ltp_osc.Oscilloscope", self._osc).start()

    @wrap_libtiepie_exception
    @_require_dev_handle(TiePieDeviceType.OSCILLOSCOPE)
    def stop_measurement(self) -> None:
        """
        Stop a measurement that is already running.

        :raises TiePieError: when device is not started, when measurement is not
        running, or when status of underlying device gives an error
        """
        if not self.is_measurement_running():
            msg = "TiePie measurement is not running"
            raise TiePieError(msg)
        cast("ltp_osc.Oscilloscope", self._osc).stop()

    @wrap_libtiepie_exception
    @_require_dev_handle(TiePieDeviceType.OSCILLOSCOPE)
    def is_measurement_running(self) -> bool:
        """
        Reports if TiePie measurement is running (ready for trigger)

        :return: if a TiePie measurement is running (ready for trigger)
        """
        _is_running = cast("ltp_osc.Oscilloscope", self._osc).is_running
        logger.debug(f"TiePie measurement is running: {_is_running}")
        return _is_running

    @wrap_libtiepie_exception
    @_require_dev_handle(TiePieDeviceType.OSCILLOSCOPE)
    def is_triggered(self) -> bool:
        """
        Reports if TiePie has triggered. Maybe data is not yet available. One can
        check with the function `is_measurement_data_ready()`.

        :return: if a trigger event occurred
        """
        _is_triggered = cast("ltp_osc.Oscilloscope", self._osc).is_triggered
        logger.debug(f"TiePie has triggered: {_is_triggered}")
        return _is_triggered

    @wrap_libtiepie_exception
    @_require_dev_handle(TiePieDeviceType.OSCILLOSCOPE)
    def is_measurement_data_ready(self) -> bool:
        """
        Reports if TiePie has data which is ready to collect

        :return: if the data is ready to collect.
        :raises TiePieError: when device is not started or status of underlying device
            gives an error
        """
        _is_measurement_data_ready = cast(
            "ltp_osc.Oscilloscope", self._osc
        ).is_data_ready
        logger.debug(f"TiePie has measurement data ready: {_is_measurement_data_ready}")
        return _is_measurement_data_ready

    @wrap_libtiepie_exception
    @_require_dev_handle(TiePieDeviceType.OSCILLOSCOPE)
    def force_trigger(self) -> None:
        """
        Forces the TiePie to trigger with a software sided trigger event.

        :return None:
        :raises TiePieError: when device is not started or status of underlying device
            gives an error
        """
        cast("ltp_osc.Oscilloscope", self._osc).force_trigger()
        logger.info("A force trigger was sent to TiePie")

    @wrap_libtiepie_exception
    @_require_dev_handle(TiePieDeviceType.OSCILLOSCOPE)
    def _check_record_length(self, data_array: npt.NDArray) -> None:
        """
        Check record length

        :param data_array: raw data from TiePie as np.ndarray, which is already
            filtered
        :return None:
        """
        if self.config_osc is None:
            logger.warning("Oscilloscope is not configured")
            return

        record_length_actual = len(data_array[:, 0])
        if record_length_actual < self.config_osc.record_length:
            logger.warning(
                "Less Data than expected: Most likely the trigger occurred "
                "before all pre trigger samples could be recorded. "
                "(pre_sample_ratio was too high)"
            )
        pre_sample_count = cast(
            "ltp_osc.Oscilloscope", self._osc
        ).valid_pre_sample_count
        record_length_predicted = (
            int(self.config_osc.record_length * (1 - self.config_osc.pre_sample_ratio))
            + pre_sample_count
        )
        if record_length_actual != record_length_predicted:
            logger.warning(
                f"The actual record length ({record_length_actual} Sa) "
                "is shorter than the "
                f"predicted record length ({record_length_predicted} Sa)"
            )

    @wrap_libtiepie_exception
    @_require_dev_handle(TiePieDeviceType.OSCILLOSCOPE)
    def collect_measurement_data(
        self, timeout: Number | None = 0
    ) -> npt.NDArray | None:
        """
        Try to collect the data from TiePie; return `None` if data is not ready.

        :param timeout: The timeout to wait until data is available. This
            option makes this function blocking the code. `timeout = None` blocks the
            code infinitely till data will be available. Per default, the `timeout`
            is set to `0`: The function will not block.
        :return: Measurement data of only enabled channels and time vector in a
            2D-`numpy.ndarray` with float sample data; or None if there is no data
            available.
        """
        # make mypy happy: config_osc could be None, which has no attributes
        if self.config_osc is None:
            logger.warning("Oscilloscope is not configured")
            return None

        if timeout is not None and not isinstance(timeout, float | int):
            msg = (
                "timeout must be non-negative number, "
                f"but '{timeout}' of type {type(timeout)} was given"
            )
            logger.error(msg)
            raise ValueError(msg)

        # Wait till timeout or till data is ready
        start_time = time.time()
        while not self.is_measurement_data_ready() and (
            timeout is None or (time.time() - start_time < timeout)
        ):
            time.sleep(self.config.is_data_ready_polling_interval_sec)

        if not self.is_measurement_data_ready():
            logger.warning(
                "Data from TiePie was not ready to collect "
                f"during a timeout of {timeout} s."
            )
            return None

        # Collect raw data from tiepie
        data = cast("ltp_osc.Oscilloscope", self._osc).get_data()
        # filter-out disabled channels entries
        data_array: npt.NDArray = np.array(list(filter(None, data))).T

        self._check_record_length(data_array)

        pre_sample_count = cast(
            "ltp_osc.Oscilloscope", self._osc
        ).valid_pre_sample_count
        record_length_actual = len(data_array[:, 0])
        time_vector = np.arange(
            -pre_sample_count, record_length_actual - pre_sample_count
        )
        time_vector = time_vector / self.config_osc.sample_rate

        return np.column_stack([time_vector, data_array])

    @property
    @_require_dev_handle(TiePieDeviceType.OSCILLOSCOPE)
    @wrap_libtiepie_exception
    def n_channels(self) -> int:
        """
        Number of channels in the oscilloscope.

        :return: Number of channels.
        """
        return len(self._osc.channels)  # type: ignore[union-attr]

    @property
    def channels_enabled(self) -> Generator[int, None, None]:
        """
        Yield numbers of enabled channels.

        :return: Numbers of enabled channels
        """
        for ch_nr, ch_config in self.config_osc_channel_dict.items():
            if ch_config.enabled:
                yield ch_nr



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Device class for controlling a Lauda PRO RP245E, circulation chiller over TCP.
"""

from .lauda import (  # noqa: F401
    LaudaProRp245e,
    LaudaProRp245eCommand,
    LaudaProRp245eCommandError,
    LaudaProRp245eConfig,
    LaudaProRp245eTcpCommunication,
    LaudaProRp245eTcpCommunicationConfig,
)



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Device class for Newport SMC100PP stepper motor controller with serial communication.

The SMC100PP is a single axis motion controller/driver for stepper motors up to 48 VDC
at 1.5 A rms. Up to 31 controllers can be networked through the internal RS-485
communication link.

Manufacturer homepage:
https://www.newport.com/f/smc100-single-axis-dc-or-stepper-motion-controller
"""

import logging
from time import sleep, time
from typing import TypeAlias, cast

# Note: PyCharm does not recognize the dependency correctly, it is added as pyserial.
from aenum import Enum, IntEnum

from hvl_ccb.comm.serial import (
    SerialCommunication,
    SerialCommunicationBytesize,
    SerialCommunicationConfig,
    SerialCommunicationParity,
    SerialCommunicationStopbits,
)
from hvl_ccb.configuration import configdataclass
from hvl_ccb.dev.base import DeviceError, SingleCommDevice
from hvl_ccb.utils.enum import AutoNumberNameEnum, NameEnum
from hvl_ccb.utils.typing import Number

logger = logging.getLogger(__name__)


Param: TypeAlias = Number | str | None


@configdataclass
class NewportSMC100PPSerialCommunicationConfig(SerialCommunicationConfig):
    #: Baudrate for NewportSMC100 controller is 57600 baud
    baudrate: int = 57600

    #: NewportSMC100 controller does not use parity
    parity: str | SerialCommunicationParity = SerialCommunicationParity.NONE

    #: NewportSMC100 controller uses one stop bit
    stopbits: int | SerialCommunicationStopbits = SerialCommunicationStopbits.ONE

    #: NewportSMC100 controller uses 8 bits for one data byte
    bytesize: int | SerialCommunicationBytesize = SerialCommunicationBytesize.EIGHTBITS

    #: The terminator is CR/LF
    terminator: bytes = b"\r\n"

    #: use 10 seconds timeout as default
    timeout: Number = 10
    #: use ASCII as de-/encoding, cf. the manual
    encoding: str = "ascii"

    #: replace bytes with � instead of raising utf-8 exception when decoding fails
    encoding_error_handling: str = "replace"


class NewportSMC100PPSerialCommunication(SerialCommunication):
    """
    Specific communication protocol implementation for NewportSMC100 controller.
    Already predefines device-specific protocol parameters in config.
    """

    class ControllerErrors(Enum, init="value message"):  # type: ignore[call-arg]
        """
        Possible controller errors with values as returned by the device in response
        to sent commands.
        """

        NO_ERROR = "@", "No error."
        CODE_OR_ADDR_INVALID = (
            "A",
            "Unknown message code or floating point controller address.",
        )
        ADDR_INCORRECT = "B", "Controller address not correct."
        PARAM_MISSING_OR_INVALID = "C", "Parameter missing or out of range."
        CMD_NOT_ALLOWED = "D", "Command not allowed."
        HOME_STARTED = "E", "Home sequence already started."
        ESP_STAGE_NAME_INVALID = (
            "F",
            "ESP stage name unknown.",
        )
        DISPLACEMENT_OUT_OF_LIMIT = (
            "G",
            "Displacement out of limits.",
        )
        CMD_NOT_ALLOWED_NOT_REFERENCED = (
            "H",
            "Command not allowed in NOT REFERENCED state.",
        )
        CMD_NOT_ALLOWED_CONFIGURATION = (
            "I",
            "Command not allowed in CONFIGURATION state.",
        )
        CMD_NOT_ALLOWED_DISABLE = "J", "Command not allowed in DISABLE state."
        CMD_NOT_ALLOWED_READY = "K", "Command not allowed in READY state."
        CMD_NOT_ALLOWED_HOMING = "L", "Command not allowed in HOMING state."
        CMD_NOT_ALLOWED_MOVING = "M", "Command not allowed in MOVING state."
        POSITION_OUT_OF_LIMIT = "N", "Current position out of software limit."
        COM_TIMEOUT = (
            "S",
            "Communication Time Out.",
        )
        EEPROM_ACCESS_ERROR = "U", "Error during EEPROM access."
        CMD_EXEC_ERROR = "V", "Error during command execution."
        CMD_NOT_ALLOWED_PP = "W", "Command not allowed for PP version."
        CMD_NOT_ALLOWED_CC = "X", "Command not allowed for CC version."

    def __init__(self, configuration) -> None:
        """
        Constructor for NewportSMC100PPSerialCommunication.
        """

        super().__init__(configuration)

    @staticmethod
    def config_cls():
        return NewportSMC100PPSerialCommunicationConfig

    def read_text(self) -> str:
        """
        Read one line of text from the serial port, and check for presence of a null
        char which indicates that the motor power supply was cut and then restored. The
        input buffer may hold additional data afterwards, since only one line is read.

        This method uses `self.access_lock` to ensure thread-safety.

        :return: String read from the serial port; `''` if there was nothing to read.
        :raises SerialCommunicationIOError: when communication port is not opened
        :raises NewportMotorPowerSupplyWasCutError: if a null char is read
        """

        with self.access_lock:
            line = super().read_text()
            if "\x00" in line:
                msg = "Unexpected message from motor:"
                raise NewportMotorPowerSupplyWasCutError(msg, line)
            return line

    def _send_command_without_checking_error(
        self, add: int, cmd: str, param: Param = None
    ) -> None:
        """
        Send a command to the controller.

        :param add: the controller address (1 to 31)
        :param cmd: the command to be sent
        :param param: optional parameter (int/float/str) appended to the command
        """

        if param is None:
            param = ""

        with self.access_lock:
            self.write_text(f"{add}{cmd}{param}")
            logger.debug(f"sent: {add}{cmd}{param}")

    def _query_without_checking_errors(
        self, add: int, cmd: str, param: Param = None
    ) -> str:
        """
        Send a command to the controller and read the answer. The prefix add+cmd is
        removed from the answer.

        :param add: the controller address (1 to 31)
        :param cmd: the command to be sent
        :param param: optional parameter (int/float/str) appended to the command
        :return: the answer from the device without the prefix
        :raises SerialCommunicationIOError: if the com is closed
        :raises NewportSerialCommunicationError: if an unexpected answer is obtained
        :raises NewportControllerError: if the controller reports an error
        """

        if param is None:
            param = ""

        prefix = f"{add}{cmd}"
        query = f"{add}{cmd}{param}"

        with self.access_lock:
            self._send_command_without_checking_error(add, cmd, param)
            sleep(0.01)
            answer = self.read_text().strip()
            if len(answer) == 0:
                message = f"Newport controller {add} did not answer to query {query}."
                logger.error(message)
                raise NewportSerialCommunicationError(message)
            if not answer.startswith(prefix):
                message = (
                    f"Newport controller {add} answer {answer} to query {query} "
                    f"does not start with expected prefix {prefix}."
                )
                logger.error(message)
                raise NewportSerialCommunicationError(message)

            logger.debug(f"Newport com: {answer}")
            return answer[len(prefix) :].strip()

    def check_for_error(self, add: int) -> None:
        """
        Ask the Newport controller for the last error it recorded.

        This method is called after every command or query.

        :param add: controller address (1 to 31)
        :raises SerialCommunicationIOError: if the com is closed
        :raises NewportSerialCommunicationError: if an unexpected answer is obtained
        :raises NewportControllerError: if the controller reports an error
        """

        with self.access_lock:
            error = self.ControllerErrors(
                self._query_without_checking_errors(add, "TE")
            )
            if error is not self.ControllerErrors.NO_ERROR:
                logger.error(f"NewportControllerError: {error.message}")
                raise NewportControllerError(error.message)

    def send_command(self, add: int, cmd: str, param: Param = None) -> None:
        """
        Send a command to the controller, and check for errors.

        :param add: the controller address (1 to 31)
        :param cmd: the command to be sent
        :param param: optional parameter (int/float/str) appended to the command
        :raises SerialCommunicationIOError: if the com is closed
        :raises NewportSerialCommunicationError: if an unexpected answer is obtained
        :raises NewportControllerError: if the controller reports an error
        """

        if param is None:
            param = ""

        with self.access_lock:
            self._send_command_without_checking_error(add, cmd, param)
            self.check_for_error(add)

    def send_stop(self, add: int) -> None:
        """
        Send the general stop ST command to the controller, and check for errors.

        :param add: the controller address (1 to 31)
        :return: ControllerErrors reported by Newport Controller
        :raises SerialCommunicationIOError: if the com is closed
        :raises NewportSerialCommunicationError: if an unexpected answer is obtained
        """

        with self.access_lock:
            self.write_text("ST")
            self.check_for_error(add)

    def query(self, add: int, cmd: str, param: Param = None) -> str:
        """
        Send a query to the controller, read the answer, and check for errors. The
        prefix add+cmd is removed from the answer.

        :param add: the controller address (1 to 31)
        :param cmd: the command to be sent
        :param param: optional parameter (int/float/str) appended to the command
        :return: the answer from the device without the prefix
        :raises SerialCommunicationIOError: if the com is closed
        :raises NewportSerialCommunicationError: if an unexpected answer is obtained
        :raises NewportControllerError: if the controller reports an error
        """

        with self.access_lock:
            try:
                answer = self._query_without_checking_errors(add, cmd, param)
            finally:
                self.check_for_error(add)

        return answer

    def query_multiple(self, add: int, cmd: str, prefixes: list[str]) -> list[str]:
        """
        Send a query to the controller, read the answers, and check for errors. The
        prefixes are removed from the answers.

        :param add: the controller address (1 to 31)
        :param cmd: the command to be sent
        :param prefixes: prefixes of each line expected in the answer
        :return: list of answers from the device without prefix
        :raises SerialCommunicationIOError: if the com is closed
        :raises NewportSerialCommunicationError: if an unexpected answer is obtained
        :raises NewportControllerError: if the controller reports an error
        """

        with self.access_lock:
            try:
                self._send_command_without_checking_error(add, cmd)
                answer = []
                for prefix in prefixes:
                    line = self.read_text().strip()
                    if not line.startswith(prefix):
                        message = (
                            f"Newport controller {add} answer {line} to command "
                            f"{cmd} does not start with expected prefix {prefix}."
                        )
                        logger.error(message)
                        raise NewportSerialCommunicationError(message)

                    answer.append(line[len(prefix) :])
            finally:
                self.check_for_error(add)

        return answer


class NewportConfigCommands(NameEnum):
    """
    Commands predefined by the communication protocol of the SMC100PP
    """

    AC = "acceleration"
    BA = "backlash_compensation"
    BH = "hysteresis_compensation"
    FRM = "micro_step_per_full_step_factor"
    FRS = "motion_distance_per_full_step"
    HT = "home_search_type"
    JR = "jerk_time"
    OH = "home_search_velocity"
    OT = "home_search_timeout"
    QIL = "peak_output_current_limit"
    SA = "rs485_address"
    SL = "negative_software_limit"
    SR = "positive_software_limit"
    VA = "velocity"
    VB = "base_velocity"
    ZX = "stage_configuration"


@configdataclass
class NewportSMC100PPConfig:
    """
    Configuration dataclass for the Newport motor controller SMC100PP.
    """

    class HomeSearch(IntEnum):
        """
        Different methods for the motor to search its home position during
        initialization.
        """

        HomeSwitch_and_Index = 0
        CurrentPosition = 1
        HomeSwitch = 2
        EndOfRunSwitch_and_Index = 3
        EndOfRunSwitch = 4

    class EspStageConfig(IntEnum):
        """
        Different configurations to check or not the motor configuration upon power-up.
        """

        DisableEspStageCheck = 1
        UpdateEspStageInfo = 2
        EnableEspStageCheck = 3

    # The following parameters are added for convenience, they do not correspond to any
    # actual hardware configuration:

    # controller address (1 to 31)
    address: int = 1

    # user position offset (mm). For convenience of the user, the motor
    # position is given relative to this point:
    user_position_offset: Number = 23.987

    # correction for the scaling between screw turns and distance (should be close to 1)
    screw_scaling: Number = 1

    # nr of seconds to wait after exit configuration command has been issued
    exit_configuration_wait_sec: Number = 5

    # waiting time for a move
    move_wait_sec: Number = 1

    # The following parameters are actual hardware configuration parameters:

    # acceleration (preset units/s^2)
    acceleration: Number = 10

    # backlash compensation (preset units)
    # either backlash compensation or hysteresis compensation can be used, not both.
    backlash_compensation: Number = 0

    # hysteresis compensation (preset units)
    # either backlash compensation or hysteresis compensation can be used, not both.
    hysteresis_compensation: Number = 0.015

    # micro step per full step factor, integer between 1 and 2000
    micro_step_per_full_step_factor: int = 100

    # motion distance per full step (preset units)
    motion_distance_per_full_step: Number = 0.01

    # home search type
    home_search_type: int | HomeSearch = HomeSearch.HomeSwitch

    # jerk time (s) -> time to reach the needed acceleration
    jerk_time: Number = 0.04

    # home search velocity (preset units/s)
    home_search_velocity: Number = 4

    # home search time-out (s)
    home_search_timeout: Number = 27.5

    # home search polling interval (s)
    home_search_polling_interval: Number = 1

    # peak output current delivered to the motor (A)
    peak_output_current_limit: Number = 0.4

    # RS485 address, integer between 2 and 31
    rs485_address: int = 2

    # lower limit for the motor position (mm)
    negative_software_limit: Number = -23.5

    # upper limit for the motor position (mm)
    positive_software_limit: Number = 25

    # maximum velocity (preset units/s), this is also the default velocity unless a
    # lower value is set
    velocity: Number = 4

    # profile generator base velocity (preset units/s)
    base_velocity: Number = 0

    # ESP stage configuration
    stage_configuration: int | EspStageConfig = EspStageConfig.EnableEspStageCheck

    def clean_values(self) -> None:
        if self.address not in range(1, 32):
            msg = "Address should be an integer between 1 and 31."
            raise ValueError(msg)
        if abs(self.screw_scaling - 1) > 0.1:
            msg = "The screw scaling should be close to 1."
            raise ValueError(msg)
        if self.exit_configuration_wait_sec <= 0:
            msg = (
                "The exit configuration wait time must be a positive "
                "value (in seconds)."
            )
            raise ValueError(msg)
        if self.move_wait_sec <= 0:
            msg = (
                "The wait time for a move to finish must be a "
                "positive value (in seconds)."
            )
            raise ValueError(msg)
        if not 1e-6 < self.acceleration < 1e12:
            msg = "The acceleration should be between 1e-6 and 1e12."
            raise ValueError(msg)
        if not 0 <= self.backlash_compensation < 1e12:
            msg = "The backlash compensation should be between 0 and 1e12."
            raise ValueError(msg)
        if not 0 <= self.hysteresis_compensation < 1e12:
            msg = "The hysteresis compensation should be between 0 and 1e12."
            raise ValueError(msg)
        if (
            not isinstance(self.micro_step_per_full_step_factor, int)
            or not 1 <= self.micro_step_per_full_step_factor <= 2000
        ):
            msg = "The micro step per full step factor should be between 1 and 2000."
            raise ValueError(msg)
        if not 1e-6 < self.motion_distance_per_full_step < 1e12:
            msg = "The motion distance per full step should be between 1e-6 and 1e12."
            raise ValueError(msg)
        if not isinstance(self.home_search_type, self.HomeSearch):
            self.force_value("home_search_type", self.HomeSearch(self.home_search_type))  # type: ignore[attr-defined]
        if not 1e-3 < self.jerk_time < 1e12:
            msg = "The jerk time should be between 1e-3 and 1e12."
            raise ValueError(msg)
        if not 1e-6 < self.home_search_velocity < 1e12:
            msg = "The home search velocity should be between 1e-6 and 1e12."
            raise ValueError(msg)
        if not 1 < self.home_search_timeout < 1e3:
            msg = "The home search timeout should be between 1 and 1e3."
            raise ValueError(msg)
        if self.home_search_polling_interval <= 0:
            msg = (
                "The home search polling interval (sec) needs to have a positive value."
            )
            raise ValueError(msg)
        if not 0.05 <= self.peak_output_current_limit <= 3:
            msg = "The peak output current limit should be between 0.05 Aand 3 A."
            raise ValueError(msg)
        if self.rs485_address not in range(2, 32):
            msg = "The RS485 address should be between 2 and 31."
            raise ValueError(msg)
        if not -1e12 < self.negative_software_limit <= 0:
            msg = "The negative software limit should be between -1e12 and 0."
            raise ValueError(msg)
        if not 0 <= self.positive_software_limit < 1e12:
            msg = "The positive software limit should be between 0 and 1e12."
            raise ValueError(msg)
        if not 1e-6 < self.velocity < 1e12:
            msg = "The velocity should be between 1e-6 and 1e12."
            raise ValueError(msg)
        if not 0 <= self.base_velocity <= self.velocity:
            msg = "The base velocity should be between 0 and the maximum velocity."
            raise ValueError(msg)
        if not isinstance(self.stage_configuration, self.EspStageConfig):
            self.force_value(  # type: ignore[attr-defined]
                "stage_configuration", self.EspStageConfig(self.stage_configuration)
            )

    def _build_motor_config(self) -> dict[str, float]:
        return {
            param.value: float(getattr(self, param.value))
            for param in NewportConfigCommands  # type: ignore[attr-defined]
        }

    @property
    def motor_config(self) -> dict[str, float]:
        """
        Gather the configuration parameters of the motor into a dictionary.

        :return: dict containing the configuration parameters of the motor
        """

        if not hasattr(self, "_motor_config"):
            self.force_value(  # type: ignore[attr-defined]
                "_motor_config",
                self._build_motor_config(),
            )
        return self._motor_config  # type: ignore[attr-defined]

    def post_force_value(self, fieldname, value) -> None:
        # if motor config is already cached and field is one of config commands fields..
        if hasattr(self, "_motor_config") and fieldname in self._motor_config:
            # ..update directly config dict value
            self._motor_config[fieldname] = value


class NewportStates(AutoNumberNameEnum):
    """
    States of the Newport controller. Certain commands are allowed only in certain
    states.
    """

    NO_REF = ()
    HOMING = ()
    CONFIG = ()
    READY = ()
    MOVING = ()
    DISABLE = ()
    JOGGING = ()


class NewportSMC100PP(SingleCommDevice):
    """
    Device class of the Newport motor controller SMC100PP
    """

    States = NewportStates

    class MotorErrors(Enum, init="value message"):  # type: ignore[call-arg]
        """
        Possible motor errors reported by the motor during get_state().
        """

        OUTPUT_POWER_EXCEEDED = 2, "80W output power exceeded"
        DC_VOLTAGE_TOO_LOW = 3, "DC voltage too low"
        WRONG_ESP_STAGE = 4, "Wrong ESP stage"
        HOMING_TIMEOUT = 5, "Homing timeout"
        FOLLOWING_ERROR = 6, "Following error"
        SHORT_CIRCUIT = 7, "Short circuit detection"
        RMS_CURRENT_LIMIT = 8, "RMS current limit"
        PEAK_CURRENT_LIMIT = 9, "Peak current limit"
        POS_END_OF_TURN = 10, "Positive end of turn"
        NED_END_OF_TURN = 11, "Negative end of turn"

    class StateMessages(Enum, init="value message state"):  # type: ignore[call-arg]
        """
        Possible messages returned by the controller on get_state() query.
        """

        NO_REF_FROM_RESET = "0A", "NOT REFERENCED from reset.", NewportStates.NO_REF
        NO_REF_FROM_HOMING = "0B", "NOT REFERENCED from HOMING.", NewportStates.NO_REF
        NO_REF_FROM_CONFIG = (
            "0C",
            "NOT REFERENCED from CONFIGURATION.",
            NewportStates.NO_REF,
        )
        NO_REF_FROM_DISABLED = (
            "0D",
            "NOT REFERENCED from DISABLE.",
            NewportStates.NO_REF,
        )
        NO_REF_FROM_READY = "0E", "NOT REFERENCED from READY.", NewportStates.NO_REF
        NO_REF_FROM_MOVING = "0F", "NOT REFERENCED from MOVING.", NewportStates.NO_REF
        NO_REF_ESP_STAGE_ERROR = (
            "10",
            "NOT REFERENCED ESP stage error.",
            NewportStates.NO_REF,
        )
        NO_REF_FROM_JOGGING = "11", "NOT REFERENCED from JOGGING.", NewportStates.NO_REF
        CONFIG = "14", "CONFIGURATION.", NewportStates.CONFIG
        HOMING_FROM_RS232 = (
            "1E",
            "HOMING commanded from RS-232-C.",
            NewportStates.HOMING,
        )
        HOMING_FROM_SMC = "1F", "HOMING commanded by SMC-RC.", NewportStates.HOMING
        MOVING = "28", "MOVING.", NewportStates.MOVING
        READY_FROM_HOMING = "32", "READY from HOMING.", NewportStates.READY
        READY_FROM_MOVING = "33", "READY from MOVING.", NewportStates.READY
        READY_FROM_DISABLE = "34", "READY from DISABLE.", NewportStates.READY
        READY_FROM_JOGGING = "35", "READY from JOGGING.", NewportStates.READY
        DISABLE_FROM_READY = "3C", "DISABLE from READY.", NewportStates.DISABLE
        DISABLE_FROM_MOVING = "3D", "DISABLE from MOVING.", NewportStates.DISABLE
        DISABLE_FROM_JOGGING = "3E", "DISABLE from JOGGING.", NewportStates.DISABLE
        JOGGING_FROM_READY = "46", "JOGGING from READY.", NewportStates.JOGGING
        JOGGING_FROM_DISABLE = "47", "JOGGING from DISABLE.", NewportStates.JOGGING

    def __init__(self, com, dev_config=None) -> None:
        # Call superclass constructor
        super().__init__(com, dev_config)

        # address of the controller
        self.address = self.config.address

        # State of the controller (see state diagram in manual)
        self.state = self.States.NO_REF

        # position of the motor
        self.position: float | None = None

    def __repr__(self) -> str:
        return f"Newport motor controller SMC100PP {self.address}"

    @staticmethod
    def default_com_cls():
        return NewportSMC100PPSerialCommunication

    @staticmethod
    def config_cls():
        return NewportSMC100PPConfig

    def start(self) -> None:
        """
        Opens the communication protocol and applies the config.

        :raises SerialCommunicationIOError: when communication port cannot be opened
        """

        logger.info(f"Starting {self}")
        super().start()

        self.get_state()

        if self.config.motor_config != self.get_motor_configuration():
            logger.info(f"Updating {self} configuration")
            if self.state != self.States.NO_REF:
                self.reset()
            self.go_to_configuration()
            self.set_motor_configuration()
            self.exit_configuration()

        if self.state == self.States.NO_REF:
            self.initialize()
            self.wait_until_motor_initialized()

    def stop(self) -> None:
        """
        Stop the device. Close the communication protocol.
        """

        try:
            if self.com.is_open:
                self.stop_motion()
        finally:
            logger.info(f"Stopping {self}")
            # close the com
            super().stop()

    def get_state(self, add: int | None = None) -> "StateMessages":
        """
        Check on the motor errors and the controller state

        :param add: controller address (1 to 31)
        :raises SerialCommunicationIOError: if the com is closed
        :raises NewportSerialCommunicationError: if an unexpected answer is obtained
        :raises NewportControllerError: if the controller reports an error
        :raises NewportMotorError: if the motor reports an error
        :return: state message from the device (member of StateMessages)
        """

        if add is None:
            add = self.address

        # sometimes when restarting the Newport controller
        # random bytes can be sent. In this scenario
        # an error is raised and the query is repeated.
        # So far on the second try the query is successful.
        # cf. Issue #93 https://gitlab.com/ethz_hvl/hvl_ccb/-/issues/93
        # and Issue #184 https://gitlab.com/ethz_hvl/hvl_ccb/-/issues/184
        try:
            ans = self.com.query(add, "TS")
        except NewportMotorPowerSupplyWasCutError:
            logger.exception("Error: Newport Motor Power Supply was Cut")
            # simply try again once
            ans = self.com.query(add, "TS")

        # the first symbol is not used, the next 3 symbols
        # are hexadecimal. Once converted to binary, they
        # indicate motor errors (see manual).
        errors = []
        for i in range(3):
            bin_errors = bin(int(ans[i + 1], 16))[2:].zfill(4)
            for j, b in enumerate(bin_errors):
                if b == "1":
                    errors.append(self.MotorErrors(i * 4 + j).message)
        if len(errors) > 0:
            message = f"Motor {add} error(s): {', '.join(errors)}"
            logger.error(message)
            raise NewportMotorError(message)
        # the next two symbols indicate the controller state
        s = self.StateMessages(ans[4:6])
        logger.info(f"The newport controller {add} is in state {s.name}")
        self.state = s.state
        return s

    def get_motor_configuration(self, add: int | None = None) -> dict[str, float]:
        """
        Query the motor configuration and returns it in a dictionary.

        :param add: controller address (1 to 31)
        :return: dictionary containing the motor's configuration
        :raises SerialCommunicationIOError: if the com is closed
        :raises NewportSerialCommunicationError: if an unexpected answer is obtained
        :raises NewportControllerError: if the controller reports an error
        """

        if add is None:
            add = self.address

        # The controller answer should be lines starting with the following prefixes
        prefixes = (
            [f"{add}PW1", f"{add}ID"]
            + [f"{add}{p.name}" for p in NewportConfigCommands]  # type: ignore[attr-defined]
            + [f"{add}PW0"]
        )
        answers = self.com.query_multiple(add, "ZT", prefixes)
        # first and last line are expected to be only the prefixes
        if answers[0] + answers[-1]:
            err_msg = "first and last line are expected to be only the prefixes"
            logger.error(err_msg)
            raise NewportError(err_msg)
        # additionally, second line ID is not relevant
        answers = answers[2:-1]

        motor_config = {}
        # for each config param, read the answer given by the controller
        for prefix, answer in zip(NewportConfigCommands, answers, strict=False):
            # cast the config param as a float and add the result to the config dict
            motor_config[prefix.value] = float(answer)
        return motor_config

    def go_to_configuration(self, add: int | None = None) -> None:
        """
        This method is executed during start(). It can also be executed after a reset().
        The controller is put in CONFIG state, where configuration parameters
        can be changed.

        :param add: controller address (1 to 31)
        :raises SerialCommunicationIOError: if the com is closed
        :raises NewportSerialCommunicationError: if an unexpected answer is obtained
        :raises NewportControllerError: if the controller reports an error
        """

        if add is None:
            add = self.address

        logger.info(f"Newport controller {add} entering CONFIG state.")
        self.com.send_command(add, "PW", 1)
        self.state = self.States.CONFIG

    def set_motor_configuration(
        self, add: int | None = None, config: dict | None = None
    ) -> None:
        """
        Set the motor configuration. The motor must be in CONFIG state.

        :param add: controller address (1 to 31)
        :param config: dictionary containing the motor's configuration
        :raises SerialCommunicationIOError: if the com is closed
        :raises NewportSerialCommunicationError: if an unexpected answer is obtained
        :raises NewportControllerError: if the controller reports an error
        """

        if add is None:
            add = self.address
        if config is None:
            config = self.config.motor_config

        logger.info(f"Setting motor {add} configuration.")
        for param in config:
            if "compensation" in param and config[param] == 0:
                logger.debug(
                    f"Skipping command to set {param} to 0, which would cause"
                    "ControllerErrors.PARAM_MISSING_OR_INVALID error. "
                    f"{param} will be set to 0 automatically anyway."
                )
            else:
                cmd = NewportConfigCommands(param).name
                self.com.send_command(add, cmd, config[param])

    def exit_configuration(self, add: int | None = None) -> None:
        """
        Exit the CONFIGURATION state and go back to the NOT REFERENCED state. All
        configuration parameters are saved to the device"s memory.

        :param add: controller address (1 to 31)
        :raises SerialCommunicationIOError: if the com is closed
        :raises NewportSerialCommunicationError: if an unexpected answer is obtained
        :raises NewportControllerError: if the controller reports an error
        """

        if add is None:
            add = self.address

        logger.info(f"Newport controller {add} leaving CONFIG state.")
        with self.com.access_lock:
            self.com._send_command_without_checking_error(add, "PW", 0)
            sleep(self.config.exit_configuration_wait_sec)
            self.com.check_for_error(add)
        self.state = self.States.NO_REF

    def initialize(self, add: int | None = None) -> None:
        """
        Puts the controller from the NOT_REF state to the READY state.
        Sends the motor to its "home" position.

        :param add: controller address (1 to 31)
        :raises SerialCommunicationIOError: if the com is closed
        :raises NewportSerialCommunicationError: if an unexpected answer is obtained
        :raises NewportControllerError: if the controller reports an error
        """

        if add is None:
            add = self.address

        logger.info(f"Newport controller {add} is HOMING.")
        self.com.send_command(add, "OR")
        self.state = self.States.READY

    def wait_until_motor_initialized(self, add: int | None = None) -> None:
        """
        Wait until the motor leaves the HOMING state (at which point it should
        have arrived to the home position).

        :param add: controller address (1 to 31)
        :raises SerialCommunicationIOError: if the com is closed
        :raises NewportSerialCommunicationError: if an unexpected answer is obtained
        :raises NewportControllerError: if the controller reports an error
        """

        if add is None:
            add = self.address

        poll = True
        elapsed_time = 0.0
        start_time = time()
        while poll:
            state_message = self.get_state(add)
            elapsed_time += time() - start_time
            poll = (state_message.state == self.States.HOMING) and (
                elapsed_time < self.config.home_search_timeout
            )
            if poll:
                sleep(self.config.home_search_polling_interval)

        if state_message != self.StateMessages.READY_FROM_HOMING:
            msg = (
                f"Newport motor {add} should be READY from"
                f" HOMING but is {state_message}."
            )
            raise NewportControllerError(msg)

    def reset(self, add: int | None = None) -> None:
        """
        Resets the controller, equivalent to a power-up. This puts the controller
        back to NOT REFERENCED state, which is necessary for configuring the controller.

        :param add: controller address (1 to 31)
        :raises SerialCommunicationIOError: if the com is closed
        :raises NewportSerialCommunicationError: if an unexpected answer is obtained
        :raises NewportControllerError: if the controller reports an error
        """

        if add is None:
            add = self.address

        logger.info(f"Newport controller {add} is being reset to NO_REF.")
        self.com.send_command(add, "RS")
        # an additional read_text is needed to clean the buffer after reset()
        strange_char = self.com.read_text()
        logger.debug(f"{self} sent this: '{strange_char}' after reset()")
        self.state = self.States.NO_REF

    def get_position(self, add: int | None = None) -> float:
        """
        Returns the value of the current position.

        :param add: controller address (1 to 31)
        :raises SerialCommunicationIOError: if the com is closed
        :raises NewportSerialCommunicationError: if an unexpected answer is obtained
        :raises NewportControllerError: if the controller reports an error
        :raises NewportUncertainPositionError: if the position is ambiguous
        """

        if add is None:
            add = self.address

        ans = float(self.com.query(add, "TP"))

        # if zero, check motor state (answer 0 is not reliable in NO_REF state)
        if ans == 0 and self.get_state().state == NewportStates.NO_REF:
            message = (
                "Motor claiming to be at home position in NO_REF state"
                "is not reliable. Initialization needed."
            )
            logger.error(message)
            raise NewportUncertainPositionError(message)

        self.position = (
            ans * self.config.screw_scaling + self.config.user_position_offset
        )
        logger.info(f"Newport motor {add} position is {self.position}.")
        return cast("float", self.position)

    def move_to_absolute_position(self, pos: Number, add: int | None = None) -> None:
        """
        Move the motor to the specified position.

        :param pos: target absolute position (affected by the configured offset)
        :param add: controller address (1 to 31), defaults to self.address
        :raises SerialCommunicationIOError: if the com is closed
        :raises NewportSerialCommunicationError: if an unexpected answer is obtained
        :raises NewportControllerError: if the controller reports an error
        """

        if add is None:
            add = self.address

        logger.info(
            f"Newport motor {add} moving from absolute position "
            f"{self.get_position()} to absolute position {pos}."
        )

        # translate user-position into hardware-position
        hard_pos = pos - self.config.user_position_offset
        self.com.send_command(add, "PA", hard_pos)
        sleep(self.config.move_wait_sec)

    def go_home(self, add: int | None = None) -> None:
        """
        Move the motor to its home position.

        :param add: controller address (1 to 31), defaults to self.address
        :raises SerialCommunicationIOError: if the com is closed
        :raises NewportSerialCommunicationError: if an unexpected answer is obtained
        :raises NewportControllerError: if the controller reports an error
        """

        if add is None:
            add = self.address

        logger.info(
            f"Newport motor {add} moving from absolute position {self.get_position()} "
            f"to home position {self.config.user_position_offset}."
        )

        self.com.send_command(add, "PA", 0)
        sleep(self.config.move_wait_sec)

    def move_to_relative_position(self, pos: Number, add: int | None = None) -> None:
        """
        Move the motor of the specified distance.

        :param pos: distance to travel (the sign gives the direction)
        :param add: controller address (1 to 31), defaults to self.address
        :raises SerialCommunicationIOError: if the com is closed
        :raises NewportSerialCommunicationError: if an unexpected answer is obtained
        :raises NewportControllerError: if the controller reports an error
        """

        if add is None:
            add = self.address

        logger.info(f"Newport motor {add} moving of {pos} units.")
        self.com.send_command(add, "PR", pos)
        sleep(self.config.move_wait_sec)

    def get_move_duration(self, dist: Number, add: int | None = None) -> float:
        """
        Estimate the time necessary to move the motor of the specified distance.

        :param dist: distance to travel
        :param add: controller address (1 to 31), defaults to self.address
        :raises SerialCommunicationIOError: if the com is closed
        :raises NewportSerialCommunicationError: if an unexpected answer is obtained
        :raises NewportControllerError: if the controller reports an error
        """

        if add is None:
            add = self.address

        dist = round(dist, 2)
        duration = float(self.com.query(add, "PT", abs(dist)))
        logger.info(f"Newport motor {add} will need {duration}s to move {dist} units.")
        return duration

    def stop_motion(self, add: int | None = None) -> None:
        """
        Stop a move in progress by decelerating the positioner immediately with the
        configured acceleration until it stops. If a controller address is provided,
        stops a move in progress on this controller, else stops the moves on all
        controllers.

        :param add: controller address (1 to 31)
        :raises SerialCommunicationIOError: if the com is closed
        :raises NewportSerialCommunicationError: if an unexpected answer is obtained
        :raises NewportControllerError: if the controller reports an error
        """

        if add is None:
            add = self.address
            logger.info("Stopping motion of all Newport motors.")
            self.com.send_stop(add)
        else:
            logger.info(f"Stopping motion of Newport motor {add}.")
            self.com.send_command(add, "ST")

    def get_acceleration(self, add: int | None = None) -> Number:
        """
        Leave the configuration state. The configuration parameters are saved to
        the device"s memory.

        :param add: controller address (1 to 31)
        :return: acceleration (preset units/s^2), value between 1e-6 and 1e12
        :raises SerialCommunicationIOError: if the com is closed
        :raises NewportSerialCommunicationError: if an unexpected answer is obtained
        :raises NewportControllerError: if the controller reports an error
        """

        if add is None:
            add = self.address

        acc = float(self.com.query(add, "AC", "?"))
        logger.info(f"Newport motor {add} acceleration is {acc}.")
        return acc

    def set_acceleration(self, acc: Number, add: int | None = None) -> None:
        """
        Leave the configuration state. The configuration parameters are saved to
        the device"s memory.

        :param acc: acceleration (preset units/s^2), value between 1e-6 and 1e12
        :param add: controller address (1 to 31)
        :raises SerialCommunicationIOError: if the com is closed
        :raises NewportSerialCommunicationError: if an unexpected answer is obtained
        :raises NewportControllerError: if the controller reports an error
        """

        if add is None:
            add = self.address

        self.com.send_command(add, "AC", acc)
        logger.info(f"Newport motor {add} acceleration set to {acc}.")

    def get_controller_information(self, add: int | None = None) -> str:
        """
        Get information on the controller name and driver version

        :param add: controller address (1 to 31)
        :return: controller information
        :raises SerialCommunicationIOError: if the com is closed
        :raises NewportSerialCommunicationError: if an unexpected answer is obtained
        :raises NewportControllerError: if the controller reports an error
        """

        if add is None:
            add = self.address

        return self.com.query(add, "VE", "?")

    def get_positive_software_limit(self, add: int | None = None) -> Number:
        """
        Get the positive software limit (the maximum position that the motor is allowed
        to travel to towards the right).

        :param add: controller address (1 to 31)
        :return: positive software limit (preset units), value between 0 and 1e12
        :raises SerialCommunicationIOError: if the com is closed
        :raises NewportSerialCommunicationError: if an unexpected answer is obtained
        :raises NewportControllerError: if the controller reports an error
        """

        if add is None:
            add = self.address

        lim = float(self.com.query(add, "SR", "?"))
        logger.info(f"Newport motor {add} positive software limit is {lim}.")
        return lim

    def set_positive_software_limit(self, lim: Number, add: int | None = None) -> None:
        """
        Set the positive software limit (the maximum position that the motor is allowed
        to travel to towards the right).

        :param lim: positive software limit (preset units), value between 0 and 1e12
        :param add: controller address (1 to 31)
        :raises SerialCommunicationIOError: if the com is closed
        :raises NewportSerialCommunicationError: if an unexpected answer is obtained
        :raises NewportControllerError: if the controller reports an error
        """

        if add is None:
            add = self.address

        self.com.send_command(add, "SR", lim)
        logger.info(f"Newport {add} positive software limit set to {lim}.")

    def get_negative_software_limit(self, add: int | None = None) -> Number:
        """
        Get the negative software limit (the maximum position that the motor is allowed
        to travel to towards the left).

        :param add: controller address (1 to 31)
        :return: negative software limit (preset units), value between -1e12 and 0
        :raises SerialCommunicationIOError: if the com is closed
        :raises NewportSerialCommunicationError: if an unexpected answer is obtained
        :raises NewportControllerError: if the controller reports an error
        """

        if add is None:
            add = self.address

        lim = float(self.com.query(add, "SL", "?"))
        logger.info(f"Newport motor {add} negative software limit is {lim}.")
        return lim

    def set_negative_software_limit(self, lim: Number, add: int | None = None) -> None:
        """
        Set the negative software limit (the maximum position that the motor is allowed
        to travel to towards the left).

        :param lim: negative software limit (preset units), value between -1e12 and 0
        :param add: controller address (1 to 31)
        :raises SerialCommunicationIOError: if the com is closed
        :raises NewportSerialCommunicationError: if an unexpected answer is obtained
        :raises NewportControllerError: if the controller reports an error
        """

        if add is None:
            add = self.address

        self.com.send_command(add, "SL", lim)
        logger.info(f"Newport {add} negative software limit set to {lim}.")


class NewportError(DeviceError):
    """
    General Error for Newport Device
    """


class NewportMotorError(NewportError):
    """
    Error with the Newport motor.
    """


class NewportUncertainPositionError(NewportError):
    """
    Error with the position of the Newport motor.
    """


class NewportMotorPowerSupplyWasCutError(NewportError):
    """
    Error with the Newport motor after the power supply was cut and then restored,
    without interrupting the communication with the controller.
    """


class NewportControllerError(NewportError):
    """
    Error with the Newport controller.
    """


class NewportSerialCommunicationError(NewportError):
    """
    Communication error with the Newport controller.
    """



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
TCP communication protocol. Makes use of the socket library.
"""

import logging
import socket
import warnings
from ipaddress import IPv4Address, IPv6Address
from typing import cast

from hvl_ccb.configuration import configdataclass
from hvl_ccb.utils.validation import (
    validate_and_resolve_host,
    validate_number,
    validate_tcp_port,
)

from .base import (
    AsyncCommunicationProtocol,
    AsyncCommunicationProtocolConfig,
    CommunicationError,
)

logger = logging.getLogger(__name__)


class TcpCommunicationError(CommunicationError):
    """Error of the TcpCommunication"""


@configdataclass
class TcpCommunicationConfig(AsyncCommunicationProtocolConfig):
    """
    Configuration dataclass for :class:`TcpCommunication`.
    """

    # Host is the IP address of the connected device.
    host: str | IPv4Address | IPv6Address | None = None
    # TCP port
    port: int = 54321
    # TCP receiving buffersize
    bufsize: int = 1024
    # TCP timeout
    timeout: float | None = 0.2

    def clean_values(self) -> None:
        # if necessary, converts host to a valid IP address
        super().clean_values()
        # Host
        self.force_value("host", validate_and_resolve_host(self.host, logger))  # type: ignore[attr-defined]
        # Port
        validate_tcp_port(self.port, logger)
        # Buffer size
        validate_number("bufsize", self.bufsize, (1, None), int, logger=logger)
        # Timeout
        validate_number("timeout", self.timeout, (0, None), logger=logger)


class TcpCommunication(AsyncCommunicationProtocol):
    """
    Tcp Communication Protocol.
    """

    config: TcpCommunicationConfig

    def __init__(self, configuration) -> None:
        """Constructor socket"""
        super().__init__(configuration)

        # create the communication port specified in the configuration
        logger.debug(
            "Create socket TcpClient with host: "
            f'"{self.config.host}", Port: "{self.config.port}"'
        )
        self._sock: socket.socket | None = None

    @staticmethod
    def config_cls() -> type[TcpCommunicationConfig]:
        return TcpCommunicationConfig

    @property
    def is_open(self) -> bool:
        """
        Is the connection open?

        :return: True for an open connection
        """
        return self._sock is not None

    def open(self) -> None:
        """
        Open TCP connection.
        """
        if self.is_open:
            logger.warning(
                "Tried to open the TCP Communication, but it is already open"
            )
            return

        # open the port
        logger.debug("Open TCP Port.")

        with self.access_lock:
            self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._sock.connect((self.config.host, self.config.port))
            self._sock.settimeout(self.config.timeout)

    def close(self) -> None:
        """
        Close TCP connection.
        """
        if not self.is_open:
            logger.warning(
                "Tried to close the TCP Communication, but it is already closed"
            )
            return

        # close the port
        logger.debug("Close TCP Port.")

        with self.access_lock:
            sock: socket.socket = cast("socket.socket", self._sock)
            sock.close()
            self._sock = None

    def write_bytes(self, data: bytes) -> int:
        """
        Write data as `bytes` to the TcpCommunication.

        This method uses `self.access_lock` to ensure thread-safety.

        :param data: data as `bytes`-string to be written
        :return: number of bytes written
        """
        if not self.is_open:
            msg = "Tried to write to the TCP Communication, but it is closed"
            logger.error(msg)
            raise TcpCommunicationError(msg)

        with self.access_lock:
            sock: socket.socket = cast("socket.socket", self._sock)
            return sock.send(data)

    def read_bytes(self) -> bytes:
        """
        Read a single line as `bytes` from the TcpCommunication.

        This method uses `self.access_lock` to ensure thread-safety.

        :return: a single line as `bytes` containing the terminator, which can also be
            empty b""
        """
        if not self.is_open:
            msg = "Tried to read from the TCP Communication, but it is closed"
            logger.error(msg)
            raise TcpCommunicationError(msg)

        with self.access_lock:
            sock: socket.socket = cast("socket.socket", self._sock)
            try:
                return sock.recv(self.config.bufsize)
            except TimeoutError:
                msg = "Timeout during reading from TCP, return empty bytes"
                logger.debug(msg)
                return b""


class Tcp(TcpCommunication):
    """Old name :class:`Tcp` for :class:`TcpCommunication` for keeping this import
    backwards compatibile"""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        warnings.warn(
            "The 'Tcp' class is deprecated, use 'TcpCommunication' instead",
            DeprecationWarning,
            stacklevel=2,
        )



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
A PICube is a BaseCube with build in Power Inverter
"""

from __future__ import annotations

import logging
from dataclasses import dataclass
from time import sleep, time

from hvl_ccb.configuration import configdataclass
from hvl_ccb.utils.typing import Number  # noqa: TC001
from hvl_ccb.utils.validation import validate_number

from . import constants
from .base import (
    BaseCube,
    BaseCubeConfiguration,
    BaseCubeOpcUaCommunication,
    BaseCubeOpcUaCommunicationConfig,
)
from .constants import _CubeOpcEndpoint
from .errors import PICubeTestParameterError

logger = logging.getLogger(__name__)


@configdataclass
class PICubeOpcUaCommunicationConfig(BaseCubeOpcUaCommunicationConfig):
    endpoint_name: _CubeOpcEndpoint = _CubeOpcEndpoint.PI_CUBE  # type: ignore[assignment]


class PICubeOpcUaCommunication(BaseCubeOpcUaCommunication):
    @staticmethod
    def config_cls():
        return PICubeOpcUaCommunicationConfig


@configdataclass
class PICubeConfiguration(BaseCubeConfiguration):
    timeout_test_parameters: Number = 2.0

    def clean_values(self) -> None:
        super().clean_values()
        if self.timeout_test_parameters < 0:
            msg = "Timeout for setting test parameters needs to be not negative."
            raise ValueError(msg)


@dataclass
class _TestParameter:
    """
    Class to access the test parameter. For the PICube it is possible to set
    a voltage target and specify the slope of the voltage ramp.
    """

    _handle: PICube
    tolerance: Number = 0.01  # tolerance is needed because of numerical precision

    @property
    def voltage(self) -> float:
        """
        Target voltage of the PICube.

        :return: target voltage in volt
        """
        display_value = float(self._handle.read(constants._Power.VOLTAGE_TARGET))
        power_setup = self._handle._power_setup
        logger.info(
            f"Test parameter: voltage is {display_value:_.2f} {power_setup.unit}"
        )
        return self._handle._apply_polarity(display_value / power_setup.scale)

    @voltage.setter
    def voltage(self, value: Number) -> None:
        """
        Target voltage of the PICube

        :param value: target voltage in volt
        :raises PICubeTestParameterError: if the value cannot be set within
            the given time
        """
        power_setup = self.power_setup
        if power_setup is None:
            return

        unit = power_setup.unit

        if self._handle.power_setup in constants.DC_POWER_SETUPS:
            if self._handle._polarity == constants.Polarity.POSITIVE:
                validate_number(
                    "voltage", value, (0, self._handle.voltage_max), logger=logger
                )
            else:
                validate_number(
                    "voltage", value, (self._handle.voltage_max, 0), logger=logger
                )
        else:
            validate_number(
                "voltage", value, (0, self._handle.voltage_max), logger=logger
            )

        voltage_opc = value * power_setup.scale

        self._handle.write(constants._Power.VOLTAGE_TARGET, abs(voltage_opc))

        start_time = time()
        rec_target_voltage = self.voltage
        while not self._check_tolerance(value, rec_target_voltage, self.tolerance):
            if (time() - start_time) >= self._handle.config.timeout_test_parameters:
                msg = (
                    f"Test parameters: voltage should be set to {value:_.2f} V, "
                    f"but is {rec_target_voltage:_.2f} V."
                )
                logger.error(msg)
                raise PICubeTestParameterError(msg)
            sleep(0.1)
            rec_target_voltage = self.voltage

        logger.info(f"Test parameters: voltage is set to {voltage_opc:_.2f} {unit}")

    @property
    def slope(self) -> float:
        """
        slope of the voltage ramp. While the value is always positive, the slope will
        be negative if the target voltage is below the current voltage.

        :return: slope in V/s
        """
        display_value = float(self._handle.read(constants._Power.VOLTAGE_SLOPE))
        power_setup = self._handle._power_setup
        logger.info(f"Test parameter: slope is {display_value:_.2f} {power_setup.unit}")
        return display_value / power_setup.scale

    @slope.setter
    def slope(self, value: Number) -> None:
        """
        slope of the voltage ramp. While the value is always positive, the slope will
        be negative if the target voltage is below the current voltage.

        :param value: slope in V/s
        :raises PICubeTestParameterError: if the value cannot be set within
            the given time
        """
        power_setup = self.power_setup
        if power_setup is None:
            return

        slope_max = power_setup.slope_max
        slope_min = power_setup.slope_min
        unit = power_setup.unit

        validate_number("slope", value, (slope_min, slope_max), logger=logger)

        slope_opc = value * power_setup.scale

        self._handle.write(constants._Power.VOLTAGE_SLOPE, slope_opc)

        start_time = time()
        rec_slope = self.slope  # tolerance is needed because of numerical precision
        while not (
            self._check_tolerance(
                value, rec_slope, self.tolerance, ignore_polarity=True
            )
        ):
            if (time() - start_time) >= self._handle.config.timeout_test_parameters:
                msg = (
                    f"Test parameters: slope should be set to {value:_.2f} V/s, "
                    f"but is {rec_slope:_.2f} V/s."
                )
                logger.error(msg)
                raise PICubeTestParameterError(msg)
            sleep(0.1)
            rec_slope = self.slope

        logger.info(f"Test parameters: slope is set to {slope_opc:_.2f} {unit}/s")

    @property
    def power_setup(self) -> constants.PowerSetup | None:
        """
        Checks if cube is in correct state and returns the power state.

        :return: Power Setup of PICube.
        """
        if self._handle._status is not constants.SafetyStatus.RED_OPERATE:
            logger.warning(
                "To set test parameters, the PICube needs to be in "
                f"Status 'RED_OPERATE', but is in '{self._handle._status.name}'"
            )
            return None
        power_setup = self._handle._power_setup
        if power_setup in (
            constants.PowerSetup.NO_SOURCE,
            constants.PowerSetup.EXTERNAL_SOURCE,
        ):
            logger.warning(
                "It is not possible to set new test parameters while "
                f"the PICube is in '{power_setup}' mode."
            )
            return None
        return power_setup

    def _check_tolerance(
        self,
        set_value: Number,
        actual_value: Number,
        tol: Number,
        ignore_polarity: bool = False,
    ) -> bool:
        """
        Checks if the actual value is within the tolerance of the set value.

        :param set_value:
        :param actual_value:
        :param tol: value between 0 and 1
        :param ignore_polarity: if `True`, always compare positive values, cf. #291
        :return: `True` if within the tolerance, otherwise `False`
        """
        if (
            self._handle._polarity == constants.Polarity.NEGATIVE
            and not ignore_polarity
        ):
            value = (1 - tol) * set_value >= actual_value >= (1 + tol) * set_value
        else:
            value = (1 - tol) * set_value <= actual_value <= (1 + tol) * set_value
        return value


class PICube(BaseCube):
    """
    Variant of the BaseCube with build in Power Inverter
    """

    def __init__(self, com, dev_config=None) -> None:
        super().__init__(com, dev_config)
        self.test_parameter = _TestParameter(self)

    @staticmethod
    def default_com_cls():
        return PICubeOpcUaCommunication

    @staticmethod
    def config_cls():
        return PICubeConfiguration

    @property
    def operate(self) -> bool | None:
        # this construct is necessary to keep mypy happy
        return super().operate

    @operate.setter
    def operate(self, state: bool) -> None:
        """
        Set operate state. If the state is RED_READY, this will turn on the high
        voltage and close the safety switches.
        If the state is RED_OPERATE and a status change to RED_READY is queried - with
        a measured output voltage above the voltage limit -- a warning is logged

        :param state: set operate state
        :raises CubeStatusChangeError: if `state=True` and cube is not in RED_READY or
            if `state=False` and cube is not in RED_OPERATE
        """

        if self.power_setup in constants.DC_POWER_SETUPS:
            voltage_limit = self.measurement_ch_3.noise_level
        elif self.power_setup in constants.AC_POWER_SETUPS:
            voltage_limit = self.measurement_ch_1.noise_level
        elif self.power_setup == constants.PowerSetup.POWER_INVERTER_220V:
            # TODO: voltage_limit for PowerInverter_220,  # noqa: FIX002
            # what is the return of actual voltage?
            # Which value is returned?
            # https://gitlab.com/ethz_hvl/hvl_ccb/-/issues/375
            voltage_limit = 10
        else:
            voltage_limit = 0

        if (
            self.voltage_actual > voltage_limit
            and self._status is constants.SafetyStatus.RED_OPERATE
            and not state
        ):
            logger.warning(
                "The output voltage needs to be 0 V in order to change "
                "Status from RED_OPERATE to RED_READY."
            )
            return
        BaseCube.operate.fset(self, state)  # type: ignore[attr-defined]

    @property
    def voltage_max(self) -> float:
        """
        Reads the maximum voltage of the setup and returns in V.

        :return: the maximum voltage of the setup in V.
        """
        power_setup = self._power_setup
        value = float(self.read(constants._Power.VOLTAGE_MAX))
        logger.info(
            "Maximum Output Voltage with current Setup "
            f"is {value:_.2f} {power_setup.unit}"
        )
        value = value / power_setup.scale
        return self._apply_polarity(value)

    @property
    def voltage_actual(self) -> float:
        """
        Reads the actual measured voltage and returns the value in V.

        :return: the actual voltage of the setup in V.
        """
        power_setup = self._power_setup
        value = float(self.read(constants._Power.VOLTAGE_ACTUAL))
        logger.info(f"Actual Output Voltage is {value:_.2f} {power_setup.unit}")
        value = value / power_setup.scale
        return self._apply_polarity(value)

    @property
    def voltage_primary(self) -> float:
        """
        Read the current primary voltage at the output of the frequency converter
        (before transformer).

        :return: primary voltage in V
        """

        value = float(self.read(constants._Power.VOLTAGE_PRIMARY))
        logger.info(
            f"Primary Voltage at the Output of the Power Inverter is {value:_.2f} V"
        )
        return value

    @property
    def current_primary(self) -> float:
        """
        Read the current primary current at the output of the frequency converter
        (before transformer).

        :return: primary current in A
        """

        value = float(self.read(constants._Power.CURRENT_PRIMARY))
        logger.info(
            f"Primary Current at the Output of the Power Inverter is {value:_.2f} A"
        )
        return value

    @property
    def frequency(self) -> float:
        """
        Read the electrical frequency of the current PICube setup.

        :return: the frequency in Hz
        """

        value = float(self.read(constants._Power.FREQUENCY))
        logger.info(f"Output Frequency of the Power Inverter is {value} Hz")
        return value

    @property
    def _power_setup(self) -> constants.PowerSetup:
        """
        Return the power setup selected in the PICube's settings, without logging

        :return: the power setup
        """
        return constants.PowerSetup(self.read(constants._Power.SETUP))

    @property
    def power_setup(self) -> constants.PowerSetup:
        """
        Return the power setup selected in the PICube's settings.

        :return: the power setup
        """

        value = self._power_setup
        logger.info(f"Current programmed Power Setup of the PICube is {value.name}")
        return value

    @property
    def _polarity(self) -> constants.Polarity | None:
        value = None
        if self._power_setup in constants.DC_POWER_SETUPS:
            value = constants.Polarity(self.read(constants._Power.POLARITY))
        return value

    @property
    def polarity(self) -> constants.Polarity | None:
        """
        Polarity of a DC setup.
        :return: if a DC setup is programmed the polarity is returned, else None.
        """

        value = self._polarity
        if value is not None:
            logger.info(f"The polarity of the experiment setup is {value.name}")
        else:
            logger.info("Only DC-Setups have a polarity.")
        return value

    def _apply_polarity(self, value: Number) -> Number:
        if (
            self._power_setup in constants.DC_POWER_SETUPS
            and self._polarity == constants.Polarity.NEGATIVE
        ):
            value = -value
        return value



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Device class for controlling a Schneider Electric ILS2T stepper drive over modbus TCP.
"""

from .se_ils2t import (  # noqa: F401
    ILS2T,
    ILS2TConfig,
    ILS2TError,
    ILS2TModbusTcpCommunication,
    ILS2TModbusTcpCommunicationConfig,
    IoScanningModeValueError,
    ScalingFactorValueError,
)



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Additional Python typing module utilities
"""

import logging
from typing import TypeAlias

import numpy.typing as npt

logger = logging.getLogger(__name__)


Number: TypeAlias = int | float
"""Typing hint auxiliary for a Python base number types: `int` or `float`."""

ConvertableTypes: TypeAlias = (
    int | float | list[Number] | tuple[Number, ...] | dict[str, Number] | npt.NDArray
)
"""Typing hint for data type that can be used in conversion"""



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Device class for controlling a Lauda PRO RP245E, circulation chiller over TCP.
"""

import logging
import time
from enum import IntEnum

from hvl_ccb.comm.tcp import TcpCommunication, TcpCommunicationConfig
from hvl_ccb.configuration import configdataclass
from hvl_ccb.dev.base import DeviceError, SingleCommDevice
from hvl_ccb.utils.enum import ValueEnum
from hvl_ccb.utils.typing import Number
from hvl_ccb.utils.validation import validate_number

logger = logging.getLogger(__name__)


class LaudaProRp245eCommand(ValueEnum):
    """
    Commands for Lauda PRO RP245E Chiller
    Command strings most often need to be complimented with a parameter
    (attached as a string) before being sent to the device.
    Commands implemented as defined in "Lauda Betriebsanleitung fuer
    PRO Badthermostate und Umwaelzthermostate" pages 42 - 49
    """

    #: Pass on external controlled temperature
    EXTERNAL_TEMP = "OUT_PV_05_"
    #: Define temperature set point
    TEMP_SET_POINT = "OUT_SP_00_"
    #: Define pump level 1-8
    PUMP_LEVEL = "OUT_SP_01_"
    #: Define operation mode
    OPERATION_MODE = "OUT_SP_02_"
    #: Define upper temp limit
    UPPER_TEMP = "OUT_SP_04_"
    #: Define lower temp limit
    LOWER_TEMP = "OUT_SP_05_"
    #: Define communication time out
    COM_TIME_OUT = "OUT_SP_08_"
    #: Set control mode 1=internal, 2=ext. analog, 3=ext. serial, 4=USB, 5=ethernet
    CONT_MODE = "OUT_MODE_01_"
    #: Start temp control (pump and heating/cooling)
    START = "START"
    #: Stop temp control (pump and heating/cooling)
    STOP = "STOP"
    #: Select a ramp program (target for all further ramp commands)
    RAMP_SELECT = "RMP_SELECT_"
    #: Start a selected ramp program
    RAMP_START = "RMP_START"
    #: Pause a selected ramp program
    RAMP_PAUSE = "RMP_PAUSE"
    #: Continue a paused ramp program
    RAMP_CONTINUE = "RMP_CONT"
    #: Stop a running ramp program
    RAMP_STOP = "RMP_STOP"
    #: Reset a selected ramp program
    RAMP_DELETE = "RMP_RESET"
    #: Define how often a ramp program should be iterated
    RAMP_ITERATIONS = "RMP_OUT_02_"
    #: Define parameters of a selected ramp program
    RAMP_SET = "RMP_OUT_00_"
    #: Request internal bath temperature
    BATH_TEMP = "IN_PV_00"
    #: Request external PT100 temperature
    EXTERNAL_PT100_TEMP = "IN_PV_03"
    #: Request device type
    DEVICE_TYPE = "TYPE"


class LaudaProRp245eError(DeviceError):
    """Errors from Lauda Devices"""


class LaudaProRp245eCommandError(LaudaProRp245eError):
    """
    Error raised when an error is returned upon a command.
    """


@configdataclass
class LaudaProRp245eTcpCommunicationConfig(TcpCommunicationConfig):
    """
    Configuration dataclass for :class:`LaudaProRp245eTcpCommunication`.
    """

    #: Delay time between commands in seconds
    wait_sec_pre_read_or_write: float = 0.005

    def clean_values(self) -> None:
        # host, raises ValueError on its own if not suitable
        super().clean_values()
        validate_number(
            "wait_sec_pre_read_or_write",
            self.wait_sec_pre_read_or_write,
            (0.001, None),
            logger=logger,
        )

        valid_terminators = {b"", b"\n", b"\r", b"\r\n"}
        if self.terminator not in valid_terminators:
            msg = f'Unknown terminator: "{self.terminator!r}"'
            logger.error(msg)
            raise ValueError(msg)


class LaudaProRp245eTcpCommunication(TcpCommunication):
    """
    Implements the Communication Protocol for Lauda PRO RP245E TCP connection.
    """

    config: LaudaProRp245eTcpCommunicationConfig

    def __init__(self, configuration) -> None:
        """Constructor for socket"""
        super().__init__(configuration)

    @staticmethod
    def config_cls() -> type[LaudaProRp245eTcpCommunicationConfig]:
        return LaudaProRp245eTcpCommunicationConfig

    def write_command(self, command: LaudaProRp245eCommand | str, param: str) -> None:
        """
        Send command function.

        :param command: first part of command string, defined in `LaudaProRp245eCommand`
        :param param: second part of command string, parameter
        :return: None
        """
        try:
            command = LaudaProRp245eCommand(command)
        except ValueError as exc:
            err_msg = f"Unknown command: '{command}///{param}'"
            logger.exception(err_msg)
            raise LaudaProRp245eCommandError(err_msg) from exc

        time.sleep(LaudaProRp245eTcpCommunicationConfig.wait_sec_pre_read_or_write)
        with self.access_lock:
            self.write(f"{command}{param}")

    def read_processed(self) -> str:
        """
        Receive value function.

        :return: reply from device as a string, the terminator, as well as the 'OK'
            stripped from the reply to make it directly useful as a value (e.g. in case
            the internal bath temperature is requested)
        """
        time.sleep(self.config.wait_sec_pre_read_or_write)
        with self.access_lock:
            return self.read().strip().replace("OK", "")

    def query_command(
        self, command: LaudaProRp245eCommand | str, param: str = ""
    ) -> str:
        """
        Send and receive function. E.g. to be used when setting/changing device setting.

        :param command: first part of command string, defined in `LaudaProRp245eCommand`
        :param param: second part of command string, parameter (by default '')
        :return: None
        """
        with self.access_lock:
            self.write_command(command, param)
            reply = self.read_processed()
        if "ERR" in reply:
            msg = f"Error in reply to a command: '{command}///{param}' => '{reply}'"
            logger.error(msg)
            raise LaudaProRp245eCommandError(msg)
        return reply

    def open(self) -> None:
        """
        Open the Lauda PRO RP245E TCP connection.
        """

        # open the port

        with self.access_lock:
            super().open()

    def close(self) -> None:
        """
        Close the Lauda PRO RP245E TCP connection.
        """

        with self.access_lock:
            # Set Lauda control mode to internal
            self.write_command(
                LaudaProRp245eCommand.CONT_MODE,
                str(LaudaProRp245eConfig.ExtControlModeEnum.INTERNAL.value),
            )
            # Stop currently running processes
            self.write_command(LaudaProRp245eCommand.STOP, "")
            # Close connection to Lauda
            super().close()


@configdataclass
class LaudaProRp245eConfig:
    """
    Configuration for the Lauda RP245E circulation chiller.
    """

    class OperationModeEnum(IntEnum):
        """Operation Mode (Cooling OFF/Cooling On/AUTO - set to AUTO)"""

        COOLOFF = 0
        COOLON = 1
        #: Automatically select heating/cooling
        AUTO = 2

    class ExtControlModeEnum(IntEnum):
        """
        Source for definition of external, controlled temperature (option 2, 3 and 4
        are not available with current configuration of the Lauda RP245E,
        add-on hardware would required)
        """

        INTERNAL = 0
        EXPT100 = 1
        ANALOG = 2
        SERIAL = 3
        USB = 4
        ETH = 5

    #: Default temperature set point
    temp_set_point_init: Number = 20.0
    #: Default pump Level
    pump_init: int = 6
    #: Upper temperature limit (safe for Galden HT135 cooling liquid)
    upper_temp: Number = 202.0
    #: Lower temperature limit (safe for Galden HT135 cooling liquid)
    lower_temp: Number = -55.0
    #: Communication time out (0 = OFF)
    com_time_out: Number = 0
    #: Highest pump level of the chiller
    max_pump_level: int = 8
    #: Maximum number of ramp programs that can be stored in the memory of the chiller
    max_pr_number: int = 5

    operation_mode: int | OperationModeEnum = OperationModeEnum.AUTO

    control_mode: int | ExtControlModeEnum = ExtControlModeEnum.INTERNAL

    def clean_values(self) -> None:
        if not isinstance(self.operation_mode, self.OperationModeEnum):
            self.force_value(  # type: ignore[attr-defined]
                "operation_mode", self.OperationModeEnum(self.operation_mode)
            )
        if not isinstance(self.control_mode, self.ExtControlModeEnum):
            self.force_value(  # type: ignore[attr-defined]
                "control_mode", self.ExtControlModeEnum(self.control_mode)
            )
        validate_number("upper_temp", self.upper_temp, (25, 202), logger=logger)
        validate_number("lower_temp", self.lower_temp, (-55, 25), logger=logger)


class LaudaProRp245e(SingleCommDevice):
    """
    Lauda RP245E circulation chiller class.
    """

    com: LaudaProRp245eTcpCommunication
    config: LaudaProRp245eConfig

    def __init__(self, com, dev_config=None) -> None:
        """
        Constructor for Lauda.

        :param com: object to use as communication protocol.
        """

        # Call superclass constructor
        super().__init__(com, dev_config)

    @staticmethod
    def default_com_cls() -> type[LaudaProRp245eTcpCommunication]:
        return LaudaProRp245eTcpCommunication

    @staticmethod
    def config_cls() -> type[LaudaProRp245eConfig]:
        return LaudaProRp245eConfig

    def _validate_pump_level(self, level: int) -> None:
        """
        Validates pump level. Raises ValueError, if pump level is incorrect.

        :param level: pump level, integer
        """
        if level > self.config.max_pump_level:
            err_msg = f"maximum pump level is {self.config.max_pump_level}"
            raise ValueError(err_msg)

    def start(self) -> None:
        """
        Start this device.

        """

        logger.info(f"Starting device {self}")

        super().start()

        self.get_device_type()
        # Defaults of things to be changed at runtime
        # safe temperature set point
        self.set_temp_set_point(self.config.temp_set_point_init)
        # standard pump level
        self.set_pump_level(self.config.pump_init)
        # Defaults set only here, not required at runtime
        self.com.query_command(
            LaudaProRp245eCommand.OPERATION_MODE, str(self.config.operation_mode)
        )
        self.com.query_command(
            LaudaProRp245eCommand.UPPER_TEMP, str(self.config.upper_temp)
        )
        self.com.query_command(
            LaudaProRp245eCommand.LOWER_TEMP, str(self.config.lower_temp)
        )
        self.com.query_command(
            LaudaProRp245eCommand.COM_TIME_OUT, str(self.config.com_time_out)
        )

    def stop(self) -> None:
        """
        Stop this device. Disables access and
        closes the communication protocol.

        """

        logger.info(f"Stopping device {self}")
        super().stop()

    def pause(self) -> str:
        """
        Stop temperature control and pump.

        :return: reply of the device to the last call of "query"
        """
        return self.com.query_command(LaudaProRp245eCommand.STOP)

    def run(self) -> str:
        """
        Start temperature control & pump.

        :return: reply of the device to the last call of "query"
        """

        return self.com.query_command(LaudaProRp245eCommand.START)

    def set_external_temp(self, external_temp: float = 20.00) -> str:
        """
        Pass value of external controlled temperature. Should be done every second,
        when control of external temperature is active. Has to be done right before
        control of external temperature is activated.

        :param external_temp: current value of external temperature to be controlled.
        :return: reply of the device to the last call of "query"
        """

        return self.com.query_command(
            LaudaProRp245eCommand.EXTERNAL_TEMP, f"{external_temp:.2f}"
        )

    def set_temp_set_point(self, temp_set_point: float = 20.00) -> str:
        """
        Define temperature set point

        :param temp_set_point: temperature set point.
        :return: reply of the device to the last call of "query"
        """

        return self.com.query_command(
            LaudaProRp245eCommand.TEMP_SET_POINT, f"{temp_set_point:.2f}"
        )

    def set_pump_level(self, pump_level: int = 6) -> str:
        """
        Set pump level
        Raises ValueError, if pump level is invalid.

        :param pump_level: pump level.
        :return: reply of the device to the last call of "query"
        """

        self._validate_pump_level(pump_level)
        return self.com.query_command(LaudaProRp245eCommand.PUMP_LEVEL, str(pump_level))

    def set_control_mode(
        self,
        mod: int
        | LaudaProRp245eConfig.ExtControlModeEnum = LaudaProRp245eConfig.ExtControlModeEnum.INTERNAL,  # noqa: E501
    ) -> str:
        """
        Define control mode. 0 = INTERNAL (control bath temp), 1 = EXPT100 (pt100
        attached to chiller), 2 = ANALOG, 3 = SERIAL, 4 = USB, 5 = ETH (to be used
        when passing the ext. temp. via ethernet) (temperature then needs to be
        passed every second, when not using options 3, 4, or 5)

        :param mod: temp control mode (control internal temp or external temp).
        :return: reply of the device to the last call of "query"
            ("OK", if command was recognized")
        """

        mod_enum = LaudaProRp245eConfig.ExtControlModeEnum(mod)
        return self.com.query_command(
            LaudaProRp245eCommand.CONT_MODE, str(mod_enum.value)
        )

    def set_ramp_program(self, program: int = 1) -> str:
        """
        Define ramp program for following ramp commands.
        Raises ValueError if maximum number of ramp programs (5) is exceeded.

        :param program: Number of ramp program to be activated for following commands.
        :return: reply of the device to the last call of "query"
        """

        if program > self.config.max_pr_number:
            err_msg = f"Maximum number of ramp programs is {self.config.max_pr_number}."
            raise ValueError(err_msg)
        return self.com.query_command(LaudaProRp245eCommand.RAMP_SELECT, str(program))

    def start_ramp(self) -> str:
        """
        Start current ramp program.

        :return: reply of the device to the last call of "query"
        """

        return self.com.query_command(LaudaProRp245eCommand.RAMP_START)

    def pause_ramp(self) -> str:
        """
        Pause current ramp program.

        :return: reply of the device to the last call of "query"
        """

        return self.com.query_command(LaudaProRp245eCommand.RAMP_PAUSE)

    def continue_ramp(self) -> str:
        """
        Continue current ramp program.

        :return: reply of the device to the last call of "query"
        """

        return self.com.query_command(LaudaProRp245eCommand.RAMP_CONTINUE)

    def stop_ramp(self) -> str:
        """
        Stop current ramp program.

        :return: reply of the device to the last call of "query"
        """

        return self.com.query_command(LaudaProRp245eCommand.RAMP_STOP)

    def reset_ramp(self) -> str:
        """
        Delete all segments from current ramp program.

        :return: reply of the device to the last call of "query"
        """

        return self.com.query_command(LaudaProRp245eCommand.RAMP_DELETE)

    def set_ramp_iterations(self, num: int = 1) -> str:
        """
        Define number of ramp program cycles.

        :param num: number of program cycles to be performed.
        :return: reply of the device to the last call of "query"
        """

        return self.com.query_command(LaudaProRp245eCommand.RAMP_ITERATIONS, str(num))

    def set_ramp_segment(
        self, temp: float = 20.00, dur: int = 0, tol: float = 0.00, pump: int = 6
    ) -> str:
        """
        Define segment of current ramp program - will be attached to current program.
        Raises ValueError, if pump level is invalid.

        :param temp: target temperature of current ramp segment
        :param dur: duration in minutes, in which target temperature should be reached
        :param tol: tolerance at which target temperature should be reached (for 0.00,
            next segment is started after dur has passed).
        :param pump: pump level to be used for this program segment.
        :return: reply of the device to the last call of "query"
        """
        self._validate_pump_level(pump)
        segment = f"{temp:.2f}_{dur}_{tol:.2f}_{pump}"
        return self.com.query_command(LaudaProRp245eCommand.RAMP_SET, segment)

    def get_bath_temp(self) -> float:
        """
        Retuns the bath temperature

        :return: float value of measured lauda bath temp in °C
        """
        rep = self.com.query_command(LaudaProRp245eCommand.BATH_TEMP)
        return float(rep)

    def get_external_temp(self) -> float:
        """
        Returns the external temperature

        :return: float value of measured external PT100 temp in °C
        """
        rep = self.com.query_command(LaudaProRp245eCommand.EXTERNAL_PT100_TEMP)
        return float(rep)

    def get_device_type(self) -> str:
        """
        Returns the device type

        :return: Connected Lauda device type (for connection/com test)
        """
        device_type = self.com.query_command(LaudaProRp245eCommand.DEVICE_TYPE)
        if "PRO" not in device_type:
            msg = "Could not connect to Lauda RP 245 E PRO. Check IP address."
            logger.error(msg)
            raise LaudaProRp245eCommandError(msg)

        return device_type



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""Devices subpackage."""

from hvl_ccb.dev.base import (  # noqa: F401
    Device,
    DeviceError,
    DeviceExistingError,
    DeviceFailuresError,
    DeviceSequenceMixin,
    SingleCommDevice,
)



#  Copyright (c) ETH Zurich, SIS ID and HVL D-ITET
#
"""
Errors of the different "Cubes".
"""

from hvl_ccb.dev import DeviceError


class CubeError(DeviceError):
    pass


class CubeStatusChangeError(CubeError):
    pass


class CubeStopError(CubeError):
    pass


class CubeRemoteControlError(CubeError):
    pass


class SwitchOperationError(CubeError):
    pass


class CubeEarthingStickOperationError(SwitchOperationError):
    pass


class DischargeError(SwitchOperationError):
    pass


class ShortCircuitError(SwitchOperationError):
    pass


class ChargerSwitchError(SwitchOperationError):
    pass


class PICubeTestParameterError(CubeError):
    pass


class AdvancedCubeModuleError(CubeError):
    pass
